# -*- coding: utf-8 -*-

from distutils.util import strtobool
from typing import Callable, Any
from xml.etree.ElementTree import ParseError

from vise.util.logger import get_logger

logger = get_logger(__name__)


def str2bool(val: str) -> bool:
    """Convert a string representation of truth to true (1) or false (0).
    """
    val = val.lower()
    if val in ('t', 'true'):
        return True
    elif val in ('false'):
        return False
    else:
        raise ValueError("invalid truth value %r" % (val,))


def is_str_digit(n: str) -> bool:
    """Check whether the given string is a digit or not. """
    try:
        float(n)
        return True
    except ValueError:
        return False


def is_str_int(n: str, rounding_error=1e-7) -> bool:
    """Check whether the given string is an integer or not. """
    try:
        if int(n) - float(n) < rounding_error:
            return True
        else:
            return False
    except ValueError:
        return False



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from math import pi, sqrt
from pathlib import Path

import pytest
from monty.serialization import loadfn
from vise.analyzer.dielectric_function import DieleFuncData, \
    eV_to_inv_cm, min_e_w_target_coeff
from vise.tests.helpers.assertion import assert_msonable, \
    assert_dataclass_almost_equal

try:
    import psutil
    PSUTIL_NOT_PRESENT = False
except ModuleNotFoundError:
    PSUTIL_NOT_PRESENT = True


@pytest.fixture
def diele_func_data():
    return DieleFuncData(energies=[0.0, 10.0],
                         directions=["xx", "yy", "ave"],
                         diele_func_real=[[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]],
                         diele_func_imag=[[1.1, 1.2], [1.3, 1.4], [1.5, 1.6]],
                         band_gap=1.0)


def test_msonable(diele_func_data):
    assert_msonable(diele_func_data)


def test_json_file_mixin(diele_func_data, tmpdir):
    tmpdir.chdir()
    diele_func_data.to_json_file()
    actual = loadfn("diele_func_data.json")
    assert actual.diele_func_real == diele_func_data.diele_func_real


def test_absorption_coeff(diele_func_data):
    actual = diele_func_data.absorption_coeff[2][1]
    real, imag = 0.6, 1.6
    expected = (2 * sqrt(2) * pi * sqrt(sqrt(real ** 2 + imag ** 2) - real)
                * 10.0 * eV_to_inv_cm)
    assert actual == expected


def test_refractive_idx(diele_func_data):
    e_real, e_imag = 0.6, 1.6  # values at xx

    actual = diele_func_data.refractive_idx_real[2][1]
    expected = sqrt(e_real + sqrt(e_real ** 2 + e_imag ** 2)) / sqrt(2)
    assert actual == expected

    actual = diele_func_data.refractive_idx_imag[2][1]
    expected = sqrt(-e_real + sqrt(e_real ** 2 + e_imag ** 2)) / sqrt(2)
    assert actual == expected


def test_reflectivity(diele_func_data):
    e_real, e_imag = 0.6, 1.6  # values at xx
    n = sqrt(e_real + sqrt(e_real ** 2 + e_imag ** 2)) / sqrt(2)
    k = sqrt(-e_real + sqrt(e_real ** 2 + e_imag ** 2)) / sqrt(2)
    expected = ((n - 1)**2 + k**2) / ((n + 1)**2 + k**2)
    actual = diele_func_data.reflectivity[2][1]
    assert actual == expected


def test_to_csv_file(tmpdir):
    diele = DieleFuncData(energies=[0.0, 3.0],
                          diele_func_real=[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]],
                          diele_func_imag=[[1.1, 1.1], [1.2, 1.2], [1.3, 1.3]],
                          directions=["xx", "yy", "ave"],
                          band_gap=1.0)
    tmpdir.chdir()
    diele.to_csv_file()
    actual = Path("diele_func_data.csv").read_text()
    expected = """energies(eV),real_xx,real_yy,real_ave,imag_xx,imag_yy,imag_ave,absorption_xx,absorption_yy,absorption_ave,band_gap
0.0,0.1,0.2,0.3,1.1,1.2,1.3,0.0,0.0,0.0,1.0
3.0,0.1,0.2,0.3,1.1,1.2,1.3,215492.69716617066,216777.74282824568,218647.7413198019,
"""
    assert actual == expected

    actual = DieleFuncData.from_csv_file("diele_func_data.csv")
    assert_dataclass_almost_equal(actual, diele, digit=3)


def test_target_coeff_e_from_band_gap():
    actual = min_e_w_target_coeff([0.0, 0.1, 0.2], [1, 2, 3], 1.5)
    assert actual == 0.1


# def test_actual_diele_func_data_with_kk_trans():
#     diele_func_data = DieleFuncData(
#         energies=[0.0, 1.0],
#         diele_func_real=[[]])
#     diele_func = make_diele_func(v, o, use_vasp_real=False)
#     print(diele_func.diele_func_real)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import numpy as np
import pytest

from vise.analyzer.dos_data import DosBySpinEnergy, DosPlotData
from vise.analyzer.plot_dos import DosPlotter, DosMplSettings
from vise.tests.helpers.assertion import assert_msonable
from vise.util.matplotlib import float_to_int_formatter

"""
TODO:
+ Create PDos class composed of orbital-decomposed pdos
+ Create from FullDos instance with total and atom- and orbital-decomposed pdos
+ Add labels
+ Allow to set vbm, cbm, efermi
+ Shift energy zero to vbm or efermi.
+ Create DosInfo from vasp output.

DONE:
"""

relative_energies = [i for i in range(-5, 6)]

total_up = [0.0] * 2 + [2] * 2 + [4] * 4 + [2] * 2 + [0]
total_down = [0.0] * 4 + [6] * 6 + [0]
h_s_up = [0.0] * 2 + [2] * 8 + [0]
h_s_down = [0.0] * 4 + [1] * 6 + [0]
h_p_up = [0.0] * 2 + [3] * 8 + [0]
h_p_down = [0.0] * 3 + [2] * 7 + [0]

xlim = [-6, 6]
ylim_set = [[-10, 10], [-5, 5]]
colors = DosMplSettings().colors


doses = [[DosBySpinEnergy("", [total_up, total_down])],
         [DosBySpinEnergy("s", np.array([h_s_up, h_s_down])),
          DosBySpinEnergy("p", np.array([h_p_up, h_p_down]))]]

dos_plot_data = DosPlotData(relative_energies=relative_energies,
                            doses=doses,
                            names=["total", "H"],
                            energy_range=xlim,
                            dos_ranges=ylim_set,
                            energy_lines=[0.0, 1.0],
                            title="test")
dos_data_len = len(dos_plot_data.doses)


def test_dos_plot_data_msonable():
    assert_msonable(dos_plot_data)


@pytest.fixture
def mock_plt_list(mocker):
    mock_plt = mocker.patch("vise.analyzer.plot_dos.plt", unsafe=True)
    mock_1st_ax = mocker.MagicMock()
    mock_2nd_ax = mocker.MagicMock()
    other_axs_len = dos_data_len - 2

    mock_axs = [mock_1st_ax, mock_2nd_ax] + [mocker.MagicMock()] * other_axs_len

    mock_plt.subplots.return_value = (None, mock_axs)
    plotter = DosPlotter(dos_plot_data=dos_plot_data)
    plotter.construct_plot()

    return mock_plt, mock_1st_ax, mock_2nd_ax


def test_dos_mpl_settings_defaults():
    mpl_settings = DosMplSettings()
    assert mpl_settings.colors == ['#36454f', '#E15759', '#4E79A7', '#F28E2B']
    assert mpl_settings.linewidth == 1.0
    assert mpl_settings.title_font_size == 15
    assert mpl_settings.label_font_size == 12
    assert mpl_settings.vline == {"linewidth": 0.75, "color": "black",
                                  "linestyle": "-."}


def test_dos_mpl_settings_modify():
    mpl_settings = DosMplSettings(colors=["black"],
                                  linewidth=2.0,
                                  band_edge_line_width=1.0,
                                  band_edge_line_color="blue",
                                  band_edge_line_style=":",
                                  title_font_size=30,
                                  label_font_size=40)

    assert mpl_settings.colors == ["black"]
    assert mpl_settings.linewidth == 2.0
    assert mpl_settings.title_font_size == 30
    assert mpl_settings.label_font_size == 40
    assert mpl_settings.vline == \
           {"linewidth": 1.0, "color": "blue", "linestyle": ":"}


def test_plot_dos(mock_plt_list):
    mock_plt, mock_1st_ax, mock_2nd_ax = mock_plt_list
    mock_plt.subplots.assert_called_once_with(dos_data_len, 1, sharex=True,
                                              gridspec_kw={'hspace': 0.1})

    reversed_total_down = [dos * -1 for dos in total_down]
    reversed_h_p_down = [dos * -1 for dos in h_p_down]

    args = DosMplSettings().dos_line(0)
    args["label"] = "total"
    mock_1st_ax.plot.assert_any_call(relative_energies, total_up, **args)

    args = DosMplSettings().dos_line(0)
    mock_1st_ax.plot.assert_any_call(
        relative_energies, reversed_total_down, **args)

    mock_1st_ax.axhline.assert_called_once_with(0, linestyle=":", color="black")

    args = DosMplSettings().dos_line(0)
    args["label"] = "H-s"
    mock_2nd_ax.plot.assert_any_call(relative_energies, h_s_up, **args)

    args = DosMplSettings().dos_line(1)
    mock_2nd_ax.plot.assert_any_call(
        relative_energies, reversed_h_p_down, **args)


def test_axs_is_list_when_single_dos_passed():
    single_dos = [[DosBySpinEnergy("", [total_up, total_down])]]
    dos_info = DosPlotData(relative_energies=relative_energies,
                           doses=single_dos, names=["total"], energy_range=xlim,
                           dos_ranges=ylim_set, energy_lines=[0.0, 1.0])
    plotter = DosPlotter(dos_plot_data=dos_info)
    assert isinstance(plotter._axs, list)
    assert len(plotter._axs) == 1


def test_set_figure_legend(mock_plt_list):
    _, mock_1st_ax, mock_2nd_ax = mock_plt_list
    mock_1st_ax.legend.assert_called_with(bbox_to_anchor=(0.9, 1),
                                          loc='upper left',
                                          borderaxespad=0,
                                          markerscale=0.1)
    mock_2nd_ax.legend.assert_called_with(bbox_to_anchor=(0.9, 1),
                                          loc='upper left',
                                          borderaxespad=0,
                                          markerscale=0.1)


def test_set_x_and_y_range(mock_plt_list):
    _, mock_1st_ax, _ = mock_plt_list
    mock_1st_ax.set_ylim.assert_called_with([-10, 10])
    mock_1st_ax.set_xlim.assert_called_with([-6, 6])


def test_set_vlines(mock_plt_list):
    _, mock_1st_ax, _ = mock_plt_list
    mpl_settings = DosMplSettings()
    mock_1st_ax.axvline(x=0.0, **mpl_settings.vline)


def test_set_labels(mock_plt_list):
    mock_plt, mock_1st_ax, mock_2nd_ax = mock_plt_list
    defaults = DosMplSettings()
    mock_plt.xlabel.assert_called_with("Energy (eV)",
                                       size=defaults.label_font_size)
    mock_1st_ax.set_ylabel.assert_called_with("Dos (1/eV)",
                                              size=defaults.label_font_size)


def test_set_float_to_int_formatter(mock_plt_list):
    _, mock_1st_ax, mock_2nd_ax = mock_plt_list
    mock_1st_ax.xaxis.set_major_formatter.assert_called_once_with(
        float_to_int_formatter)
    mock_1st_ax.yaxis.set_major_formatter.assert_called_once_with(
        float_to_int_formatter)
    mock_2nd_ax.yaxis.set_major_formatter.assert_called_once_with(
        float_to_int_formatter)


def test_actual_plot():
    plotter = DosPlotter(dos_plot_data)
    plotter.construct_plot()
    plotter.plt.show()


# def test_plotly_actual_plot():
#     pploter = PlotlyDosPlotter(dos_plot_data)
#     pploter.show()


# # 1. imports of your dash app
# import dash
# import dash_html_components as html


# # 2. give each testcase a tcid, and pass the fixture
# # as a function argument, less boilerplate
# def test_bsly001_falsy_child(dash_duo):

    # # 3. define your app inside the test function
    # app = dash.Dash(__name__)
    # app.layout = html.Div(id="nully-wrapper", children=0)

    # # 4. host the app locally in a thread, all dash server configs could be
    # # passed after the first app argument
    # dash_duo.start_server(app)

    # # 5. use wait_for_* if your target element is the result of a callback,
    # # keep in mind even the initial rendering can trigger callbacks
    # dash_duo.wait_for_text_to_equal("#nully-wrapper", "0", timeout=4)

    # # 6. use this form if its present is expected at the action point
    # assert dash_duo.find_element("#nully-wrapper").text == "0"

    # # 7. to make the checkpoint more readable, you can describe the
    # # acceptance criterion as an assert message after the comma.
    # assert dash_duo.get_logs() == [], "browser console should contain no error"

    # # 8. visual testing with percy snapshot
    # dash_duo.percy_snapshot("bsly001-layout")


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import warnings
from pathlib import Path
from typing import Union, List

import yaml
from monty.serialization import loadfn
from pymatgen.core import Structure, Lattice, Element
from pymatgen.io.vasp import Kpoints
from pymatgen.io.vasp.sets import MPRelaxSet, BadInputSetWarning
from vise.input_set.datasets.potcar_set import PotcarSet

mags = loadfn(Path(__file__).parent / "mp_atom_mag.yaml")


warnings.simplefilter('ignore', BadInputSetWarning)


def is_target_element(elem: Union[Element, str]):
    elem = Element(elem) if type(elem) is str else elem
    return elem.Z <= 58 or 71 <= elem.Z <= 83


def make_atom_poscar_dirs(path: Path, elems: List[Element] = None):
    elems = [str(e) for e in elems] if elems else mags.keys()

    for element in elems:
        d = path / element
        Path(d).mkdir()
        structure = Structure(Lattice.cubic(10), [element], [[0.5]*3])
        structure.to(fmt="POSCAR", filename=str(d / "POSCAR"))
        nupdown = mags[element]
        prior_info = {"incar": {"ISPIN": 2, "NUPDOWN": float(nupdown),
                                "NELM": 300},
                      "is_cluster": True}
        (d / "prior_info.yaml").write_text(yaml.dump(prior_info))


def make_atom_mp_relax_set():
    for element, potcar in PotcarSet.mp_relax_set.potcar_dict().items():
        if potcar is None:
            continue

        if is_target_element(element) is False:
            continue

        structure = Structure(Lattice.cubic(10),
                              coords=[[0.5]*3], species=[element])

        mp_set = MPRelaxSet(structure,
                            user_kpoints_settings=Kpoints(kpts=((1, 1, 1),)),
                            user_incar_settings={"ALGO": "D",
                                                 "ISIF": 2,
                                                 "ISMEAR": 0,
                                                 "MAGMOM": {"H": 1.0},
                                                 "NELM": 300})
        Path(element).mkdir()
        mp_set.write_input(element)


if __name__ == '__main__':
    make_atom_mp_relax_set()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import numpy as np
from pymatgen.io.vasp import Vasprun, Outcar
from vise.analyzer.dielectric_function import make_shifted_diele_func
from vise.analyzer.vasp.make_diele_func import make_diele_func, make_average


def test_make_diele_func(test_data_files):
    v = Vasprun(test_data_files / "MgSe_absorption_vasprun_gamma.xml")
    o = Outcar(test_data_files / "MgSe_absorption_OUTCAR_gamma")
    actual = make_diele_func(v, o)
#    print(VaspBandEdgeProperties(v, o))
    print(actual.diele_func_real)
    assert actual.energies[1] == 0.0407


def test_make_diele_func_calc_real(test_data_files):
    v = Vasprun(test_data_files / "MgSe_absorption_vasprun_gamma.xml")
    o = Outcar(test_data_files / "MgSe_absorption_OUTCAR_gamma")
    actual = make_diele_func(v, o, use_vasp_real=False)
    print(actual.diele_func_real)
    assert actual.energies[1] == 0.0407


def test_make_diele_func_2(test_data_files):
    v = Vasprun(test_data_files / "MgSe_absorption_vasprun_gamma.xml")
    o = Outcar(test_data_files / "MgSe_absorption_OUTCAR_gamma")
    actual = make_shifted_diele_func(make_diele_func(v, o),
                                     original_band_gap=1.997,
                                     shift=1.0)
    assert isinstance(actual.diele_func_imag[0], list)
    assert isinstance(actual.diele_func_imag[0][0], float)
#    print(type(actual.diele_func_imag[0]))
#    print(type(actual.diele_func_imag[0][0]))
    #    print(actual.diele_func_real)
    # print(actual.diele_func_imag[1000])
    # plotter = AbsorptionCoeffPlotlyPlotter(actual)
    # fig = plotter.create_figure()
    # fig.show()


def test_make_ave():
    actual = make_average(
        vals=np.array([[0.0, 1.0], [1.0, 2.0], [2.0, 3.0], [100, 200]]))
    expected = np.array([1.0, 2.0])
    assert (actual == expected).all()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
from pymatgen.core import Element

from vise.input_set.datasets.potcar_set import PotcarSet
from vise.input_set.potcar_generator import generate_potcar, ViseNoPotcarError
from vise.input_set.xc import Xc


@pytest.fixture()
def mg_symbol():
    return [str(Element("Mg"))]


@pytest.fixture()
def max_z_Cm_symbol():
    return [str(Element.from_Z(96))]


@pytest.fixture()
def Bk_symbol():
    return [str(Element.from_Z(97))]


def test_normal_mg(mg_symbol):
    potcar = generate_potcar(mg_symbol, Xc.pbe)
    expected = "  PAW_PBE Mg 13Apr2007".split()
    assert str(potcar).split("\n")[0].split() == expected


def test_mp_relax_set_mg(mg_symbol):
    potcar = generate_potcar(mg_symbol,
                             Xc.pbe,
                             potcar_set=PotcarSet.mp_relax_set)
    expected = "  PAW_PBE Mg_pv 13Apr2007".split()
    assert str(potcar).split("\n")[0].split() == expected


def test_gw(mg_symbol):
    potcar = generate_potcar(mg_symbol, Xc.pbe, potcar_set=PotcarSet.gw)
    expected = "  PAW_PBE Mg_GW 13Apr2007                    "
    assert str(potcar).split("\n")[0] == expected


def test_normal_largest_z(max_z_Cm_symbol):
    potcar = generate_potcar(max_z_Cm_symbol, Xc.pbe)
    expected = "  PAW_PBE Cm 17Jan2011                   "
    assert str(potcar).split("\n")[0] == expected


def test_not_exist(Bk_symbol):
    with pytest.raises(ViseNoPotcarError):
        generate_potcar(Bk_symbol, Xc.pbe)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import pytest
from vise.cli.main_tools import potcar_str2dict, list2dict


def test_none():
    assert potcar_str2dict(None) == {}


def test_list():
    expected = {"Mg": "Mg_pv", "O": "O_h"}
    actual = potcar_str2dict(["Mg_pv", "O_h"])
    assert actual == expected


def test_str():
    expected = {"Mg": "Mg_pv"}
    actual = potcar_str2dict("Mg_pv")
    assert actual == expected


def test_mutliple_potcars_for_same_element_error():
    with pytest.raises(ValueError):
        potcar_str2dict(["Mg_pv", "Mg"])


def test_incorrect_element_potcar_error():
    with pytest.raises(ValueError):
        potcar_str2dict(["MgHH", "Mg"])


def test_list2dict():
    flattened_list = ["ENCUT", "500", "MAGMOM", "4", "4.0", "LWAVE", "F"]
    actual = list2dict(flattened_list, ["ENCUT", "MAGMOM", "LWAVE"])
    expected = {"ENCUT": 500, "MAGMOM": [4, 4.0], "LWAVE": False}
    assert actual == expected


def test_list2dict_2():
    flattened_list = ["EFIELD_PEAD", "0.1", "0.1", "0.1"]
    actual = list2dict(flattened_list, ["EFIELD_PEAD"])
    expected = {"EFIELD_PEAD": [0.1, 0.1, 0.1]}
    assert actual == expected


def test_list2dict_3():
    flattened_list = ["ALGO", "N"]
    actual = list2dict(flattened_list, ["ALGO"])
    expected = {"ALGO": "N"}
    assert actual == expected


def test_fail():
    flattened_list = ["ENCAT", "500"]
    with pytest.raises(ValueError):
        list2dict(flattened_list, ["ENCUT"])


def test_fail2():
    flattened_list = ["ENCUT", "500", "MAGMOM"]
    with pytest.raises(ValueError):
        list2dict(flattened_list, ["ENCUT", "MAGMOM"])





# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.
from vise.util.string import numbers_to_lowercases


def test_numbers_to_lowercases():
    assert numbers_to_lowercases("Mg2") == "Mg₂"


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from math import ceil
from typing import Optional, List, Union

import numpy as np
from pymatgen.core import Structure
from pymatgen.io.vasp.sets import Kpoints
from vise.defaults import defaults
from vise.input_set.kpoints_mode import KpointsMode
from vise.input_set.task import Task
from vise.util.bravais_lattice import BravaisLattice
from vise.util.logger import get_logger
from vise.util.structure_symmetrizer import StructureSymmetrizer

logger = get_logger(__name__)


class StructureKpointsGenerator:
    def __init__(
            self,
            initial_structure: Structure,
            task: Task,
            kpt_density: float,  # in Å
            kpt_mode: Union[KpointsMode, str, None] = None,  # None for default
            gamma_centered: Optional[bool] = None,  # Vasp definition
            only_even_num_kpts: bool = False,  # If ceil kpt numbers to be even.
            num_kpt_factor: Optional[int] = None,  # Set NKRED to this as well.
            band_ref_dist: float = defaults.band_mesh_distance,
            symprec: float = defaults.symmetry_length_tolerance,
            angle_tolerance: float = defaults.symmetry_angle_tolerance,
            time_reversal: bool = True):  # Whether the system has time reversal symmetry.

        if kpt_mode:
            kpt_mode = KpointsMode.from_string(str(kpt_mode))

        self._initial_structure = initial_structure.copy()
        self._task = task
        self._kpt_density = kpt_density
        self._time_reversal = time_reversal
        self._symprec = symprec
        self._angle_tolerance = angle_tolerance
        self._num_kpt_factor = num_kpt_factor or self._task.default_kpt_factor
        if self._num_kpt_factor != 1:
            logger.info(f"kpoint factor is set to {self._num_kpt_factor}")
        self._symmetrizer = \
            StructureSymmetrizer(structure=self._initial_structure,
                                 symprec=symprec,
                                 angle_tolerance=angle_tolerance,
                                 band_mesh_distance=band_ref_dist,
                                 time_reversal=time_reversal)
        # overwrite options fixed by other options
        self._gamma_centered = task.requisite_gamma_centered or gamma_centered
        self._adjust_only_even_num_kpts(only_even_num_kpts)
        self._adjust_kpt_mode(kpt_mode)

    def _adjust_only_even_num_kpts(self, option):
        task_requisite = self._task.requisite_only_even_num_kpts
        if task_requisite:
            logger.info(f"Numbers of kpoints are incremented to be even "
                        f"numbers.")
        if isinstance(task_requisite, bool):
            result = task_requisite
        else:
            result = option
        self._only_even_num_kpts = result

    def _adjust_kpt_mode(self, option):
        self._kpt_mode = option or self._task.default_kpt_mode

    @property
    def _longest_idx(self):
        return np.argmax(self.structure.lattice.abc)

    def generate_input(self):
        self._make_structure()
        self._make_kpoints()

    def _make_structure(self):
        if self._kpt_mode == KpointsMode.uniform:
            self._structure = self._initial_structure.copy()
        elif self._kpt_mode == KpointsMode.primitive:
            self._structure = self._symmetrizer.primitive
        elif self._kpt_mode == KpointsMode.band:
            self._structure = self._symmetrizer.band_primitive
        else:
            raise NotImplementedError
        if self._structure != self._initial_structure:
            logger.info("The input structure is changed. When you do not want "
                        "to change it, use --uniform_kpt_mode.")

    def _make_kpoints(self):
        self._set_num_kpt_list()
        self._set_kpt_shift()
        self._set_kpoints()

    def _set_num_kpt_list(self) -> None:
        if self._task.requisite_num_kpt_list:
            self._num_kpt_list = self._task.requisite_num_kpt_list
            return

        kpt_list = []
        for reciprocal_lattice_length in self._reciprocal_lat_abc:
            a_kpt_num = self._kpt_density * reciprocal_lattice_length

            if self._only_even_num_kpts:
                rounded_up_kpt_num = ceil(a_kpt_num / 2) * 2
            else:
                rounded_up_kpt_num = ceil(a_kpt_num)

            kpt_list.append(rounded_up_kpt_num * self._num_kpt_factor)

        if self._task == Task.defect_2d:
            direction = "xyz"[self._longest_idx]
            logger.info(f"The number of kpoints along the {direction} "
                        f"direction is set to 1.")
            kpt_list[self._longest_idx] = 1

        self._num_kpt_list = kpt_list

    @property
    def _reciprocal_lat_abc(self) -> List[float]:
        recipro_lat_abc = self._structure.lattice.reciprocal_lattice.abc
        self.bravais = BravaisLattice.from_sg_num(self._symmetrizer.sg_number)
        if self._kpt_mode.band_or_primitive and self.bravais.need_same_num_kpt:
            logger.warning("To keep the space group symmetry, the number "
                           "of k-points along three directions must be the "
                           "same for oI and tI Bravais lattice.")
            geometric_mean_abc = pow(float(np.prod(recipro_lat_abc)), 1 / 3)
            result = (geometric_mean_abc,) * 3
        else:
            result = recipro_lat_abc

        return result

    def _set_kpt_shift(self):
        if self._gamma_centered:
            self._kpt_shift = [0.0, 0.0, 0.0]
            return

        if self._kpt_mode is KpointsMode.uniform:
            kpt_shift = []
            angles = self._structure.lattice.angles
            for i in range(3):
                normal_to_plane = max([abs(angles[i - 2] - 90),
                                       abs(angles[i - 1] - 90)]) < 1e-5
                kpt_shift.append(0.5 if normal_to_plane else 0.0)
        else:
            kpt_shift = self.bravais.kpt_centering

        if kpt_shift[self._longest_idx] != 0.0 and self._task is Task.defect_2d:
            kpt_shift[self._longest_idx] = 0.0

        even_num_kpt_pos = [self._num_kpt_list[i] % 2 == 0 for i in range(3)]
        self._kpt_shift = [s * a for s, a in zip(kpt_shift, even_num_kpt_pos)]

    def _set_kpoints(self):
        # symmetrizer must be recreated since the fractional coordinates are
        # different in different lattices.
        # Symmetrized structure doesn't need the symprec and angle_tolerance
        symmetrizer = StructureSymmetrizer(self._structure,
                                           symprec=self._symprec,
                                           angle_tolerance=self._angle_tolerance,
                                           time_reversal=self._time_reversal)
        irreducible_kpoints = symmetrizer.irreducible_kpoints(
            self._num_kpt_list, self._kpt_shift)

        self.comment = ""
        if self._kpt_mode is KpointsMode.band:
            self._add_weighted_kpts(irreducible_kpoints)
            self._add_band_path_kpts()
            self._num_kpts = len(self._kpoints.kpts)
        else:
            self._kpoints = Kpoints(comment=self.comment,
                                    kpts=(self._num_kpt_list,),
                                    kpts_shift=self._kpt_shift)
            self._num_kpts = len(irreducible_kpoints)

            self.kpoints.comment += f"Num irrep kpoints: {self._num_kpts}"

    def _add_weighted_kpts(self, irreducible_kpoints):
        kpts = [k[0] for k in irreducible_kpoints]
        weights = [k[1] for k in irreducible_kpoints]
        labels = [None] * len(irreducible_kpoints)
        self._kpoints = Kpoints(comment=self.comment,
                                style=Kpoints.supported_modes.Reciprocal,
                                num_kpts=len(kpts),
                                kpts=kpts,
                                kpts_weights=weights,
                                labels=labels)

    def _add_band_path_kpts(self):
        k_path = self._symmetrizer.seekpath_data["explicit_kpoints_rel"]
        k_labels = self._symmetrizer.seekpath_data["explicit_kpoints_labels"]
        formula = self._structure.composition.reduced_formula
        sg = self._symmetrizer.sg_number
        self.kpoints.comment += \
            f"k-path added by seekpath. Formula: {formula} SG: {sg} "
        self._kpoints.num_kpts += len(k_path)
        self._kpoints.kpts += list(k_path)
        self._kpoints.labels += k_labels
        self._kpoints.kpts_weights += [0] * len(k_path)

    @property
    def structure(self):
        return self._structure

    @property
    def kpoints(self):
        return self._kpoints

    @property
    def num_kpts(self):
        return self._num_kpts

    @property
    def num_kpt_factor(self):
        return self._num_kpt_factor




# -*- coding: utf-8 -*-

#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from collections import defaultdict

import numpy
from pymatgen.io.vasp import Vasprun, Outcar, Procar
from vise.analyzer.band_edge_properties import BandEdgeProperties


class VaspBandEdgeProperties(BandEdgeProperties):
    def __init__(self,
                 vasprun: Vasprun,
                 outcar: Outcar,
                 integer_criterion: float = 0.1):

        total_mag = outcar.total_mag if outcar.total_mag else 0.0

        self.structure = vasprun.final_structure

        super().__init__(eigenvalues=eigenvalues_from_vasprun(vasprun),
                         nelect=outcar.nelect,
                         magnetization=total_mag,
                         kpoint_coords=vasprun.actual_kpoints,
                         integer_criterion=integer_criterion,
                         is_non_collinear=
                         vasprun.parameters.get("LNONCOLLINEAR", False))


def edge_orbital_contributions(procar: Procar, structure, vbm_info, cbm_info):
    elements = structure.composition.elements
    ele_indices = []
    for e in elements:
        ele_indices.append(
            [i for i, s in enumerate(structure) if s.specie == e])

    # Kpoint index first and then band index
    vbm = procar.data[vbm_info.spin][vbm_info.kpoint_index][vbm_info.band_index]
    cbm = procar.data[cbm_info.spin][cbm_info.kpoint_index][cbm_info.band_index]

    # orbital type
    if len(vbm[0]) == 3:
        orb_indices = [[0], [1], [2]]
    elif len(vbm[0]) == 4:
        orb_indices = [[0], [1], [2], [3]]
    elif len(vbm[0]) == 9:
        orb_indices = [[0], [1, 2, 3], [4, 5, 6, 7, 8]]
    elif len(vbm[0]) == 16:
        orb_indices = [[0], [1, 2, 3], [4, 5, 6, 7, 8],
                       [9, 10, 11, 12, 13, 14, 15]]
    else:
        raise ValueError("")

    vbm_comp, cbm_comp = defaultdict(dict), defaultdict(dict)

    for e, e_idxs in zip(elements, ele_indices):
        for o, o_idxs in zip(["s", "p", "d", "f"], orb_indices):
            vbm_comp[str(e)][o] = vbm[numpy.ix_(e_idxs, o_idxs)].sum()
            cbm_comp[str(e)][o] = cbm[numpy.ix_(e_idxs, o_idxs)].sum()

    return dict(vbm_comp), dict(cbm_comp)


def eigenvalues_from_vasprun(vasprun):
    return {spin: e[:, :, 0] for spin, e in vasprun.eigenvalues.items()}


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import os
from pathlib import Path

import pytest

from vise.util.file_transfer import (
    FileTransfer, FileMove, FileCopy, FileLink, FileTransfers,
    ViseFileTransferError
)

cwd = Path.cwd().absolute()


@pytest.fixture()
def file_transfer_instance():
    class Test(FileTransfer):
        def transfer(self, abs_output_dir):
            pass

    return Test(abs_file_path=cwd / "a")


def test_file_transfer_abs_file_path(file_transfer_instance):
    assert file_transfer_instance.abs_in_file == cwd / "a"


def test_file_transfer_file_name(file_transfer_instance):
    assert file_transfer_instance.file_name == "a"


def test_file_transfer_to(file_transfer_instance):
    assert file_transfer_instance.to(cwd.absolute() / "b") == cwd / "b" / "a"


@pytest.fixture
def tmp_dirs(tmpdir):
    written_string = "written_string"
    filename = "a"
    tmp_from = tmpdir
    tmp_from.join(filename).write(written_string)
    tmp_to = tmp_from.mkdir("tmp_to")
    yield tmp_from, tmp_to, filename, written_string


def test_file_move(tmp_dirs):
    tmp_from, tmp_to, filename, written_string = tmp_dirs
    file_move = FileMove(abs_file_path=Path(tmp_from) / filename)
    file_move.transfer(Path(tmp_to))
    os.chdir(Path(tmp_to))
    with open(Path(tmp_to) / filename, 'r') as f2:
        assert f2.read() == written_string


def test_file_copy(tmp_dirs):
    tmp_from, tmp_to, filename, _ = tmp_dirs
    file_copy = FileCopy(abs_file_path=Path(tmp_from) / filename)
    file_copy.transfer(Path(tmp_to))
    os.chdir(Path(tmp_to))
    with open(Path(tmp_to) / filename, 'r') as f_to:
        with open(Path(tmp_from) / filename, 'r') as f_from:
            assert f_to.read() == f_from.read()


def test_file_link(tmp_dirs):
    tmp_from, tmp_to, filename, _ = tmp_dirs
    file_link = FileLink(abs_file_path=Path(tmp_from) / filename)
    file_link.transfer(Path(tmp_to))
    assert (Path(tmp_to) / filename).is_symlink()

    os.chdir(Path(tmp_to))
    with open(Path(tmp_to) / filename, 'r') as f_to:
        with open(Path(tmp_from) / filename, 'r') as f_from:
            assert f_to.read() == f_from.read()


def test_transfer_files(tmpdir):
    test_string = "test"
    tmp_from = tmpdir
    os.chdir(tmp_from)
    for i in ["file1", "file2", "file3", "file4"]:
        with open(i, "w") as f1:
            print(test_string, end="", file=f1)
    file_transfers = FileTransfers(
        {"file1": "m", "file2": "c", "file3": "l", "file4": "m"},
        path=Path(tmp_from))

    tmp_to = tmp_from.mkdir("tmp_to")
    file_transfers.transfer(Path(tmp_to))
    # Test copied and linked files
    for i in ["file2", "file3"]:
        with open(Path(tmp_to) / i, 'r') as f_to:
            with open(Path(tmp_from) / i, 'r') as f_from:
                assert f_to.read() == f_from.read()

    # Test moved files
    for i in ["file1", "file4"]:
        with open(Path(tmp_to) / i, 'r') as f2:
            assert f2.read() == test_string


def test_transfer_files_logger_non_exist(tmpdir, mocker):
    mock = mocker.patch("vise.util.file_transfer.logger.warning")
    non_exist_filename = "non_exist_file"
    tmp_from = tmpdir
    os.chdir(tmp_from)
    FileTransfers({non_exist_filename: "m"}, path=Path(tmp_from))
    mock.assert_called_once_with(
        f"{Path(tmp_from) / non_exist_filename} does not exist.")


def test_transfer_files_logger_empty(tmpdir, mocker):
    mock = mocker.patch("vise.util.file_transfer.logger.warning")
    empty_filename = "empty_file"
    tmp_from = tmpdir
    tmp_from.join(empty_filename).write("")
    FileTransfers({empty_filename: "m"}, path=Path(tmp_from))
    mock.assert_called_once_with(
        f"{Path(tmp_from) / empty_filename} is empty.")


def test_raise_error_for_incorrect_transfer_type(tmpdir):
    tmp_from = tmpdir
    os.chdir(tmp_from)
    tmp_from.join("filename").write("test")
    with pytest.raises(ViseFileTransferError):
        FileTransfers({"filename": "x"}, path=Path(tmp_from))


def test_transfer_delete(tmpdir):
    tmp_from = tmpdir
    for i in ["a", "bbb", "cbb", "bd", "bda"]:
        tmp_from.join(i).write("test")
    file_transfers = FileTransfers(
        {"a":   "m", "bbb": "c", "cbb": "l", "bd":  "m", "bda": "m"},
        path=Path(tmp_from))
    file_transfers.delete_file_transfers(["bb", "a"])
    assert len(file_transfers.file_transfers) == 1



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
from pymatgen.core import Composition, Structure, Element
from pymatgen.io.vasp import Potcar

from vise.input_set.incar_settings_generator import (
    IncarSettingsGenerator)
from vise.input_set.task import Task
from vise.input_set.xc import Xc
from vise.defaults import defaults


lattice = [[1, 0, 0], [2, 0, 0], [0, 0, 3]]


@pytest.fixture()
def default_dict():
    return {"structure": Structure(lattice,
                                   species=["H", "O", "O"],
                                   coords=[[1, 0, 0], [2, 1, 0], [0, 0, 3]]),
            "symbol_list": ["H", "O"],
            "num_kpts": 5,
            "num_kpt_multiplication_factor": 1,
            "potcar": Potcar(["H", "O"]),
            "xc": Xc.pbe,
            "task": Task.structure_opt}


def test_pbe_structure_opt(default_dict):
    generator = IncarSettingsGenerator(**default_dict)
    # for k, v in generator.incar_settings.items():
    #     print(f'\"{k}\": {v}, ')
    expected = {
        "ALGO": "Normal",
        "PREC": "Normal",
        "LREAL": False,
        "EDIFF": 1e-07,
        "ENCUT": 520.0,
        "LASPH": True,
        "NELM": 100,
        "ISIF": 3,
        "IBRION": 2,
        "EDIFFG": -0.005,
        "NSW": 20,
        "ISMEAR": 0,
        "SIGMA": 0.1,
        "LWAVE": False,
        "LCHARG": False,
        "LORBIT": 10,
        "KPAR": 2,
    }
    assert generator.incar_settings == expected


def test_scan_structure_opt(default_dict):
    default_dict["xc"] = Xc.scan
    generator = IncarSettingsGenerator(**default_dict)
    # for k, v in generator.incar_settings.items():
    #     print(f'\"{k}\": {v}, ')
    assert generator.incar_settings["METAGGA"] == "SCAN"


def test_hse_structure_opt(default_dict):
    default_dict.update({"xc": Xc.hse})
    generator = IncarSettingsGenerator(**default_dict)
    expected = {
        "ALGO": "Damped",
        "PREC": "Normal",
        "LREAL": False,
        "EDIFF": 1e-07,
        "ENCUT": 520.0,
        "LASPH": True,
        "NELM": 100,
        "ISIF": 3,
        "IBRION": 2,
        "EDIFFG": -0.005,
        "NSW": 20,
        "ISMEAR": 0,
        "SIGMA": 0.1,
        "LWAVE": True,
        "LCHARG": False,
        "LORBIT": 10,
        "LHFCALC": True,
        "PRECFOCK": "Fast",
        "TIME": 0.4,
        "AEXX": 0.25,
        "HFSCREEN": 0.208,
        "KPAR": 2,
    }
    assert generator.incar_settings == expected


def test_hse_structure_opt_2(default_dict):
    default_dict.update({"xc": Xc.hse,
                         "structure": Structure(
                             lattice, species=["U", "O", "O"],
                             coords=[[1, 0, 0], [2, 1, 0], [0, 0, 3]]),
                         "symbol_list": ["U", "O"],
                         "potcar": Potcar(["U", "O"]),
                         "exchange_ratio": 0.5,
                         "set_hubbard_u": True})

    generator = IncarSettingsGenerator(**default_dict)
    # for k, v in generator.incar_settings.items():
    #     print(f'\"{k}\": {v}, ')
    expected = {
        "ALGO": "Damped",
        "PREC": "Normal",
        "LREAL": False,
        "EDIFF": 1e-07,
        "ENCUT": 520.0,
        "LASPH": True,
        "NELM": 100,
        "ISIF": 3,
        "IBRION": 2,
        "EDIFFG": -0.005,
        "NSW": 20,
        "ISMEAR": 0,
        "SIGMA": 0.1,
        "LWAVE": True,
        "LCHARG": False,
        "LORBIT": 10,
        "LHFCALC": True,
        "PRECFOCK": "Fast",
        "TIME": 0.4,
        "AEXX": 0.5,
        "HFSCREEN": 0.208,
        "LDAU": True,
        "LDAUTYPE": 2,
        "LMAXMIX": 6,
        "LDAUPRINT": 1,
        "LDAUU": [5, 0],
        "LDAUL": [3, -1],
        "KPAR": 2,
    }
    assert generator.incar_settings == expected


def test_dos(default_dict):
    default_dict.update({"vbm_cbm": [0.0, 1.0], "task": Task.dos})
    generator = IncarSettingsGenerator(**default_dict)
    # for k, v in generator.incar_settings.items():
    #     print(f'\"{k}\": {v}, ')
    expected = {
        "ALGO": "Normal",
        "PREC": "Normal",
        "LREAL": False,
        "EDIFF": 1e-05,
        "ENCUT": 400.0,
        "LASPH": True,
        "NELM": 100,
        "ISIF": 0,
        "IBRION": 2,
        "NSW": 1,
        "ISMEAR": -4,
        "SIGMA": 0.1,
        "LWAVE": False,
        "LCHARG": False,
        "NBANDS": 19,
        "LORBIT": 11,
        "EMIN": -15.01,
        "EMAX": 16,
        "NEDOS": 3102,
        "KPAR": 2,
    }
    assert generator.incar_settings == expected


def test_defect(default_dict):
    default_dict.update({"task": Task.defect})
    generator = IncarSettingsGenerator(**default_dict)
    # for k, v in generator.incar_settings.items():
    #     print(f'\"{k}\": {v}, ')
    expected = {
        "ALGO": "Normal",
        "PREC": "Normal",
        "LREAL": "Auto",
        "EDIFF": 1e-05,
        "ENCUT": 400.0,
        "LASPH": True,
        "NELM": 100,
        "ISIF": 2,
        "IBRION": 2,
        "EDIFFG": -0.03,
        "NSW": 20,
        "ISMEAR": 0,
        "SIGMA": 0.1,
        "ISPIN": 2,
        "LWAVE": False,
        "LCHARG": False,
        "LORBIT": 10,
        "KPAR": 2,
    }
    assert generator.incar_settings == expected


def test_phonon_force(default_dict):
    default_dict.update({"task": Task.phonon_force})
    generator = IncarSettingsGenerator(**default_dict)
    # for k, v in generator.incar_settings.items():
    #     print(f'\"{k}\": {v}, ')
    expected = {
        "ADDGRID": True,
        "ALGO": "Normal",
        "PREC": "Accurate",
        "LREAL": False,
        "EDIFF": 1e-08,
        "ENCUT": 400.0,
        "LASPH": True,
        "NELM": 100,
        "ISIF": 2,
        "IBRION": 2,
        "NSW": 1,
        "ISMEAR": 0,
        "SIGMA": 0.1,
        "LWAVE": False,
        "LCHARG": False,
        "KPAR": 2,
    }
    assert generator.incar_settings == expected


def test_dielectric_finite_field(default_dict):
    default_dict.update({"task": Task.dielectric_finite_field})
    generator = IncarSettingsGenerator(**default_dict)
    expected = {
        'ALGO': 'Normal',
        'EDIFF': 1e-06,
        'ENCUT': 400.0,
        'IBRION': 2,
        'ISIF': 0,
        'ISMEAR': 0,
        'KPAR': 2,
        'LASPH': True,
        'LCALCEPS': True,
        'LCHARG': False,
        'LORBIT': 10,
        'LREAL': False,
        'LWAVE': False,
        'NELM': 100,
        'NSW': 1,
        'POTIM': 0.015,
        'PREC': 'Normal',
        'SIGMA': 0.1}
    assert generator.incar_settings == expected


def test_dielectric_function(default_dict):
    default_dict.update({"task": Task.dielectric_function})
    generator = IncarSettingsGenerator(**default_dict)
    # for k, v in generator.incar_settings.items():
    #     print(f'\"{k}\": {v}, ')
    expected = {
        "ALGO": "Normal",
        "PREC": "Normal",
        "LREAL": False,
        "EDIFF": 1e-05,
        "ENCUT": 400.0,
        "LASPH": True,
        "NELM": 100,
        "ISIF": 0,
        "IBRION": 2,
        "NSW": 1,
        "ISMEAR": 0,
        "SIGMA": 0.1,
        "LWAVE": False,
        "LCHARG": False,
        "NBANDS": 19,
        "LORBIT": 10,
        "EMIN": -20.01,
        "EMAX": 20,
        "NEDOS": 4002,
        "LOPTICS": True,
        "CSHIFT": 0.0,
        "KPAR": 2,
    }
    assert generator.incar_settings == expected


def test_with_band_gap_band(default_dict):
    default_dict.update({"band_gap": defaults.band_gap_criterion + 1e-5,
                         "task": Task.band})
    generator = IncarSettingsGenerator(**default_dict)
    assert generator.incar_settings["ISMEAR"] == 0


def test_with_band_gap_normal(default_dict):
    default_dict.update({"band_gap": defaults.band_gap_criterion + 1e-5})
    generator = IncarSettingsGenerator(**default_dict)
    assert generator.incar_settings["ISMEAR"] == -5


def test_with_band_gap_dielectric_func(default_dict):
    default_dict.update({"task": Task.dielectric_function,
                         "band_gap": defaults.band_gap_criterion + 1e-5})
    generator = IncarSettingsGenerator(**default_dict)
    assert generator.incar_settings["ISMEAR"] == -4


def test_ldau_option():
    generator = IncarSettingsGenerator(
        structure=Structure(lattice, Composition("Zn"), coords=[[0, 0, 0]]),
        symbol_list=["Zn"],
        potcar=Potcar(["Zn"]),
        num_kpts=5,
        num_kpt_multiplication_factor=1,
        xc=Xc.pbe,
        task=Task.structure_opt)
    assert "LDAU" not in generator.incar_settings


def test_grids_option(default_dict):
    default_dict["multiples_for_grids"] = [5, 2, 4]
    generator = IncarSettingsGenerator(**default_dict)
    assert generator.incar_settings["NGX"] == 10
    assert generator.incar_settings["NGY"] == 12
    assert generator.incar_settings["NGZ"] == 20


def test_spin_orbit(default_dict):
    default_dict["set_spin_orbit"] = True
    generator = IncarSettingsGenerator(**default_dict)
    assert generator.incar_settings["LSORBIT"] is True


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import csv
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Dict, List, Any, Tuple

import pandas as pd
import yaml
from monty.serialization import loadfn
from vise.util.logger import get_logger


logger = get_logger(__name__)


class ToFileMixIn(ABC):
    @property
    def _filename(self):
        """ ClassForThis -> class_for_this
        https://stackoverflow.com/questions/7322028/how-to-replace-uppercase-with-underscore
        """
        class_name = self.__class__.__name__
        return re.sub("(?<!^)(?=[A-Z])", "_", class_name).lower()


class ToJsonFileMixIn(ToFileMixIn, ABC):
    def to_json_file(self,
                     filename: Optional[str] = None,
                     suffix: Optional[str] = None) -> None:
        filename = (filename or self.fname_w_suffix(suffix)
                    or self.json_filename)
        Path(filename).write_text(self.to_json())

    @abstractmethod
    def to_json(self):
        pass

    @property
    def json_filename(self) -> str:
        return self._filename + ".json"

    def fname_w_suffix(self, suffix) -> str:
        if suffix:
            return f"{self._filename}_{suffix}.json"


class ToCsvFileMixIn(ToFileMixIn, ABC):

    def to_csv_file(self,
                    filename: Optional[str] = None,
                    ) -> None:
        self.to_dataframe.to_csv(filename or self._csv_filename, index=False)

        # filename = Path(filename or self._csv_filename)
        # with open(filename, 'w', newline='') as file:
        #     writer = csv.writer(file)
        #     names, data = self.csv_data
        #     writer.writerow(names)
        #     writer.writerows(data)

    @classmethod
    def from_csv_file(cls, filename):
        return cls.from_dataframe(pd.read_csv(filename))

    @property
    @abstractmethod
    def to_dataframe(self) -> pd.DataFrame:
        pass

    @classmethod
    @abstractmethod
    def from_dataframe(cls, df):
        pass

    @property
    def _csv_filename(self) -> str:
        return self._filename + ".csv"


class ToYamlFileMixIn(ToFileMixIn, ABC):

    def to_yaml_file(self, filename: Optional[str] = None) -> None:
        filename = filename or self._yaml_filename
        Path(filename).write_text(self.to_yaml())

    def to_yaml(self):
        return yaml.dump(self.as_dict())

    # This is not abstractmethod because we don't need this when to_yaml and
    # from_yaml are overridden.
    def as_dict(self):
        pass

    @classmethod
    def from_yaml(cls, filename: str):
        with open(filename) as file:
            d = yaml.safe_load(file)
        if hasattr(cls, "from_dict"):
            return cls.from_dict(d)
        return cls(**d)

    @property
    def _yaml_filename(self):
        """ ClassForThis -> class_for_this.json
        https://stackoverflow.com/questions/7322028/how-to-replace-uppercase-with-underscore
        """
        return self._filename + ".yaml"





# -*- coding: utf-8 -*-
#  Copyright (c) 2022. Distributed under the terms of the MIT License.
from typing import List, Tuple

import numpy as np
from pymatgen.core import Structure
from pymatgen.io.vasp import Kpoints

from vise.analyzer.plot_band import Irrep, Irreps
from vise.analyzer.vasp.plot_band import greek_to_unicode
from vise.error import ViseError
from vise.util.logger import get_logger
from vise.util.structure_symmetrizer import StructureSymmetrizer

logger = get_logger(__name__)


def special_points_from_kpoints(kpoints_filename: str) \
        -> Tuple[List[str], List[int]]:

    kpoints = Kpoints.from_file(kpoints_filename)
    special_points, kpt_indices = [], []

    for idx, label in enumerate(kpoints.labels, 1):
        if label in [None, "None"] or label in special_points:
            continue
        special_points.append(label)
        kpt_indices.append(idx)

    special_points = [x.replace("GAMMA", "GM") for x in special_points]

    return special_points, kpt_indices


# def make_irreps_from_wavecar(special_point_symbols: List[str],
#                              kpt_indices: List[int],
#                              sg_num: int = None,
#                              wavecar_filename: str = "WAVECAR",
#                              poscar_filename: str = "POSCAR",
#                              plane_wave_cutoff: float = 50.0,
#                              degeneracy_threshold: float = 0.01,
#                              **bs_kwargs) -> Irreps:
#
#     if sg_num is None:
#         structure = Structure.from_file(poscar_filename)
#         sg_analyzer = StructureSymmetrizer(structure=structure)
#         sg_num = sg_analyzer.sg_number
#
#     try:
#         from irreptables import IrrepTable
#         from irrep.bandstructure import BandStructure
#     except ImportError:
#         logger.warning(f"To find irreps, install the irrep package.")
#         raise
#     logger.info("We set spinor is False.")
#     irrep_table = IrrepTable(sg_num, spinor=False)
#     symbols = {i.kpname for i in irrep_table.irreps}
#
#     listed_k_indices, listed_symbols = [], []
#     for s, k_index in zip(special_point_symbols, kpt_indices):
#         if s in symbols:
#             listed_symbols.append(s)
#             listed_k_indices.append(k_index)
#         else:
#             logger.info(f"{s} is not listed in the IrrepTable.")
#
#     bs = BandStructure(fWAV=wavecar_filename,
#                        fPOS=poscar_filename,
#                        Ecut=plane_wave_cutoff,
#                        kplist=np.array(listed_k_indices),
#                        spinor=False,
#                        **bs_kwargs)
#
#     characters = bs.write_characters(degen_thresh=degeneracy_threshold,
#                                      kpnames=listed_symbols)
#
#     irrep_dict = {}
#     for symbol, c_kpt, kpt in \
#             zip(listed_symbols, characters["k-points"], bs.kpoints):
#         symbols = []
#         for irrep in c_kpt["irreps"]:
#             try:
#                 irrep_str = find_irrep(irrep)
#             except ViseNoIrrepError:
#                 irrep_str = "Unknown"
#             symbols.append(greek_to_unicode(irrep_str))
#
#         irrep_dict[greek_to_unicode(symbol)] = \
#             Irrep(kpt.K.tolist(),
#                   symbols,
#                   c_kpt["energies"].tolist(),
#                   c_kpt["dimensions"].tolist())
#
#     return Irreps(bs.spacegroup.number, irrep_dict)


class ViseNoIrrepError(ViseError):
    pass


def find_irrep(d: dict, threshold: float = 0.99):
    for k, v in d.items():
        if v[0] > threshold:
            return k
    logger.warning(f"""Any irrep could not be found. 
Threshold: {threshold}
Characters: {d}""")
    raise ViseNoIrrepError



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import os
import shutil
from abc import ABC, abstractmethod
from pathlib import Path
import re
from typing import List, Dict

from monty.io import zopen

from vise.util.logger import get_logger
from vise.error import ViseError
logger = get_logger(__name__)


class FileTransfer(ABC):
    def __init__(self, abs_file_path: Path):
        self.abs_in_file = abs_file_path

    @property
    def file_name(self):
        return Path(self.abs_in_file).name

    @abstractmethod
    def transfer(self, abs_output_dir: Path):
        pass

    def to(self, abs_output_dir: Path) -> Path:
        return abs_output_dir / self.file_name


class FileMove(FileTransfer):
    def transfer(self, abs_output_dir: Path):
        shutil.move(self.abs_in_file, self.to(abs_output_dir))


class FileCopy(FileTransfer):
    def transfer(self, abs_output_dir: Path):
        with zopen(self.abs_in_file, "rb") as fin, \
                zopen(self.to(abs_output_dir), "wb") as fout:
            shutil.copyfileobj(fin, fout)


class FileLink(FileTransfer):
    def transfer(self, abs_output_dir: Path):
        os.symlink(self.abs_in_file, self.to(abs_output_dir))


def transfer_instance(transfer_type: str, filename: Path):
    initial = transfer_type[0].lower()

    if initial == "m":
        result = FileMove
    elif initial == "c":
        result = FileCopy
    elif initial == "l":
        result = FileLink
    else:
        raise ViseFileTransferError(
            f"Transfer type with an initial of {initial} is not supported."
            f"Choose from m (move), c (copy) or l (link).")
    return result(filename)


class FileTransfers:
    def __init__(self, file_transfer_types: Dict[str, str], path: Path):

        file_transfers = []
        for filename, transfer_type in file_transfer_types.items():
            f = path.absolute() / filename

            if not f.is_file():
                logger.warning(f"{f} does not exist.")
            elif f.stat().st_size == 0:
                logger.warning(f"{f} is empty.")
            else:
                file_transfers.append(transfer_instance(transfer_type, f))

        self.file_transfers = file_transfers

    def delete_file_transfers(self, keywords: List[str]):
        pattern = re.compile("|".join(keywords))
        for file_transfer in list(self.file_transfers):
            if pattern.search(file_transfer.file_name):
                self.file_transfers.remove(file_transfer)

    def transfer(self, output_dir: Path = Path(".")) -> None:
        for transfer in self.file_transfers:
            transfer.transfer(output_dir.absolute())


class ViseFileTransferError(ViseError):
    pass



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass

from vise.util.enum import ExtendedEnum


def test_extended_enum():

    @dataclass
    class Test(ExtendedEnum):
        a: str = "a"

    assert Test("a") == Test.a
    assert Test.values() == ["a"]




# -*- coding: utf-8 -*-
#  Copyright (c) 2021. Distributed under the terms of the MIT License.

from scipy.constants import physical_constants as pc
au_to_angstrom = pc["atomic unit of length"][0] * 10 ** 10  # = 0.5291772109



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import  dataclasses
import sys
from difflib import Differ
from pathlib import Path

from monty.json import MSONable
from monty.serialization import loadfn
import numpy as np
from pymatgen.analysis.structure_matcher import StructureMatcher, \
    ElementComparator
from pymatgen.core import IStructure
from vise.util.mix_in import ToYamlFileMixIn


def assert_structure_almost_same(s1: IStructure, s2: IStructure):
    matcher = StructureMatcher(ltol=0.0001,
                               stol=0.0001,
                               angle_tol=0.0001,
                               primitive_cell=False,
                               scale=False,
                               attempt_supercell=False,
                               allow_subset=False,
                               comparator=ElementComparator())
    assert matcher.fit(s1, s2, skip_structure_reduction=True)


def assert_msonable(obj):
    assert isinstance(obj, MSONable)
    rounded_obj = obj.__class__.from_dict(obj.as_dict())
    try:

        if dataclasses.is_dataclass(obj):
            try:
                assert obj == rounded_obj
            except ValueError: # numpy array cannot be compared.
                pass
        assert obj.as_dict() == rounded_obj.as_dict()

    except AssertionError:
        print(obj)
        print(rounded_obj)
        print("compare items")
        for k, v in obj.__dict__.items():
            rounded = rounded_obj.__dict__[k]
            print(rounded)
            print(k, v, rounded, type(v), type(rounded))
            try:
                print(v == rounded)
            except ValueError:
                print((v == rounded).all())
        raise


def assert_json_roundtrip(obj, tmpdir):
    tmpdir.chdir()
    obj.to_json_file("a.json")
    actual = loadfn("a.json").as_dict()
    expected = obj.as_dict()
    for k, v in actual.items():
        try:
            if isinstance(v, np.ndarray):
                (v == expected[k]).all()
            elif isinstance(v, dict):
                for kk, vv in v.items():
                    try:
                        if isinstance(vv, np.ndarray):
                            (vv == expected[k][kk]).all()
                        else:
                             assert vv == expected[k][kk]
                    except ValueError:
                        print("v", v)
                        print(type(v))
                        print("k", expected[k])
                        print(type(expected[k]))
                        raise
            else:
                assert v == expected[k]
        except AssertionError:
            print(k)
            print(v)
            print(expected[k])
            raise
        except ValueError:
            print("v", v)
            print(type(v))
            print("k", expected[k])
            print(type(expected[k]))
            raise

def assert_yaml_roundtrip(obj: ToYamlFileMixIn,
                          tmpdir,
                          expected_text: str,
                          compare_dict: bool = True,
                          compare_items: bool = True):
    tmpdir.chdir()
    obj.to_yaml_file("a.yaml")
    actual_text = Path("a.yaml").read_text()
    try:
        assert actual_text == expected_text
    except AssertionError:
        print(tmpdir)
        print("actual text")
        print(actual_text)
        a = actual_text.split("\n")
        b = expected_text.split("\n")
        sys.stdout.writelines(list(Differ().compare(a, b)))
        raise

    if compare_dict:
        try:
            actual = obj.__class__.from_yaml("a.yaml").as_dict()
            expected = obj.as_dict()
            # When as_dict is not implemented, actual becomes None.
            if actual is None:
                raise AttributeError
        except AttributeError:
            from dataclasses import asdict
            actual = asdict(obj.from_yaml("a.yaml"))
            expected = asdict(obj)
        try:
            assert len(actual) == len(expected)
        except (AssertionError, TypeError):
            print(tmpdir)
            raise
    else:
        actual = obj.from_yaml("a.yaml")
        print("actual", actual)

    if compare_items is True:
        for k, v in actual.items():
            try:
                assert v == expected[k]
            except AssertionError:
                print(tmpdir)
                print(f"key: {k}, actual: {v}, expected: {expected[k]}")
                raise


def assert_dataclass_almost_equal(actual, expected, digit=8, check_is_subclass=False):
    assert actual.__class__ == expected.__class__

    for k, v1 in actual.__dict__.items():
        try:
            v2 = expected.__getattribute__(k)
            assert_attribute_almost_same(v1, v2, digit, check_is_subclass)
        except AssertionError:
            print(f"Error raised key {k}")
            raise


def assert_attribute_almost_same(v1, v2, digit, check_is_subclass=False):
    try:
        if check_is_subclass:
            assert issubclass(v1.__class__, v2.__class__)
        else:
            assert v1.__class__ == v2.__class__

    except AssertionError:
        print("type(v1), type(v2)")
        print(type(v1), type(v2))
        raise

    if isinstance(v1, float):
        try:
            assert round(v1, digit) == round(v2, digit)
        except AssertionError:
            print(v1, v2)
            raise
    elif isinstance(v1, tuple) or isinstance(v1, list):
        for vv1, vv2 in zip(v1, v2):
            try:
                assert_attribute_almost_same(vv1, vv2, digit, check_is_subclass)
            except AssertionError:
                print("original keys")
                print("v1", v1)
                print("v2", v2)
                raise
    elif dataclasses.is_dataclass(v1):
        try:
            assert_dataclass_almost_equal(v1, v2, digit, check_is_subclass)
        except AssertionError:
            print(v1, v2)
            raise
    else:
        try:
            if isinstance(v1, np.ndarray):
                try:
                    np.testing.assert_array_almost_equal(v1, v2)
                except AssertionError:
                    print(v1, v2)
                    raise
            else:
                try:
                    assert v1 == v2
                except AssertionError:
                    print(v1, v2)
                    raise
        except AssertionError:
            print(v1, v2)
            raise



# -*- coding: utf-8 -*-
#  Copyright (c) 2023. Distributed under the terms of the MIT License.
import numpy as np
from numpy.linalg import det
from numpy import concatenate, clip, dot, arctan2


def sort_coords(coords: np.ndarray) -> np.ndarray:
    """Sort coordinates based on the angle with first coord from the center.

    Args:
        coords (np.ndarray):
            Coordinates to be sorted. The format of coords is as follows.
            np.array([[x1, y1, z1], [x2, y2, z2], [x3, y3, z3]]

    Returns:
        np.ndarray for sorted coordinates.
    """
    if len(coords[0]) != 3:
        raise ValueError("Only valid for 3D vector")

    center = np.average(coords, axis=0)
    relative_coords = coords - center
    external_prod = np.cross(relative_coords[0], relative_coords[1])

    # Skip parallel vectors.
    if abs(np.linalg.norm(external_prod)) < 1e-8 and len(relative_coords) > 2:
        external_prod = np.cross(relative_coords[0], relative_coords[2])
    normal_to_12_plane = external_prod / np.linalg.norm(external_prod)

    v0 = relative_coords[0] / np.linalg.norm(relative_coords[0])

    def angle_between_v0(index: int) -> float:
        """
        Args:
            index (int): index of coords.

        Returns (float):
            Angle between rays from the center to rel_coords[0] and
            rel_coords[int].
        """
        v = relative_coords[index] / np.linalg.norm(relative_coords[index])
        matrix = concatenate(([v0], [v], [normal_to_12_plane]), axis=0)
        determinant = det(matrix)
        angle = arctan2(clip(dot(v0, v), -1.0, 1.0), determinant)
        return angle

    indices = [i for i in range(len(coords))]
    indices.sort(key=angle_between_v0)
    return coords[indices]



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import List
import numpy as np

from monty.json import MSONable
from numpy import linalg
from tabulate import tabulate
from vise.util.mix_in import ToJsonFileMixIn


@dataclass
class EffectiveMass(MSONable, ToJsonFileMixIn):
    p: List[List[List]]
    n: List[List[List]]
    temperature: float
    concentrations: List[float]

    def average_mass(self, carrier_type, concentration):
        em = self.effective_mass(carrier_type, concentration)
        return sum(em[i][i] for i in range(3)) / 3

    def minimum_mass(self, carrier_type, concentration):
        em = self.effective_mass(carrier_type, concentration)
        result = lowest_eigval_and_vecs(np.array(em))[0]
        if isinstance(result, complex):
            if abs(result.imag) < 1e-3:
                result = result.real
            else:
                raise ValueError("The minimum effective mass shows complex "
                                 f"value of f{result}.")
        return result

    def effective_mass(self, carrier_type, concentration):
        i_c = self.concentrations.index(concentration)
        return self.__getattribute__(carrier_type)[i_c]

    def __str__(self):
        lines = [f"temperature: {self.temperature}"]
        for c, pp, nn in zip(self.concentrations, self.p, self.n):
            lines.append("-"*30)
            lines.append(f"concentration: {c:g}")
            lines.append(f"p:")
            lines.append(tabulate(pp))
            lines.append(f"n:")
            lines.append(tabulate(nn))
        return "\n".join(lines)


def eigvals_and_vecs(matrix: np.ndarray):
    assert matrix.shape == (3, 3)
    eigvals, eigvecs = linalg.eig(matrix)
    idx = eigvals.argsort()
    eigvals = eigvals[idx]
    eigvecs = eigvecs[:, idx].T  # Need to transverse
    return eigvals, eigvecs


def lowest_eigval_and_vecs(matrix: np.ndarray):
    eigvals, eigvecs = eigvals_and_vecs(matrix)
    assert eigvals[0] > 0.0
    eigval = eigvals[0]
    lowest_eigvals = np.where(eigvals < eigval * 1.0001)
    eigvecs = np.round(eigvecs[lowest_eigvals], 2)
    return eigval, eigvecs



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest

from vise import __version__
from vise.defaults import defaults
from vise.input_set.input_options import (
    CategorizedInputOptions, ViseInputOptionsError)
from vise.input_set.task import Task
from vise.input_set.vise_log import ViseLog
from vise.input_set.xc import Xc


def test_no_options(sc_structure):
    opts = CategorizedInputOptions(sc_structure, task=Task.structure_opt, xc=Xc.pbe)

    structure_kpoints_keys = set(opts.structure_kpoints_options.keys())
    assert structure_kpoints_keys == {"kpt_density", "initial_structure",
                                      "task"}

    potcar_keys = set(opts.potcar_options.keys())
    assert potcar_keys == {"xc"}

    incar_settings_keys = set(opts.incar_settings_options.keys())
    assert incar_settings_keys == {'xc', 'task'}


def test_initial_structures(sc_structure):
    opts = CategorizedInputOptions(sc_structure, task=Task.structure_opt, xc=Xc.pbe)
    assert str(opts.initial_structure.composition) == "H1"


def test_raise_error(sc_structure):
    with pytest.raises(ViseInputOptionsError):
        CategorizedInputOptions(sc_structure, task=Task.structure_opt, xc=Xc.pbe,
                                fake_option="")


def test_structure_kpoints_options(sc_structure):
    opts = CategorizedInputOptions(sc_structure, task=Task.structure_opt,
                                   xc=Xc.pbe, kpt_density=100)

    assert opts.structure_kpoints_options["kpt_density"] == 100


def test_potcar_options(sc_structure):
    opts = CategorizedInputOptions(sc_structure, task=Task.structure_opt, xc=Xc.pbe,
                                   overridden_potcar={"A": "B"})

    assert opts.potcar_options["overridden_potcar"] == {"A": "B"}


def test_incar_settings_options(sc_structure):
    opts = CategorizedInputOptions(sc_structure, task=Task.structure_opt, xc=Xc.pbe,
                                   charge=100.0)

    assert opts.incar_settings_options["charge"] == 100.0


def test_initial_structure(sc_structure):
    opts = CategorizedInputOptions(sc_structure, task=Task.structure_opt, xc=Xc.pbe,
                                   charge=100.0)

    assert opts.initial_structure == sc_structure


def test_insulator_kpt_density(sc_structure):
    opts = CategorizedInputOptions(sc_structure, task=Task.structure_opt,
                                   xc=Xc.pbe, vbm_cbm=[0, 0.5])
    actual = opts.structure_kpoints_options["kpt_density"]
    expected = defaults.insulator_kpoint_density
    assert actual == expected

    opts = CategorizedInputOptions(sc_structure, task=Task.structure_opt,
                                   xc=Xc.pbe, band_gap=1.0)
    actual = opts.structure_kpoints_options["kpt_density"]
    expected = defaults.insulator_kpoint_density
    assert actual == expected

    opts = CategorizedInputOptions(sc_structure, task=Task.structure_opt,
                                   xc=Xc.pbe, vbm_cbm=[0, 0.5],
                                   kpt_density=100)
    assert opts.structure_kpoints_options["kpt_density"] == 100


def test_parameter_dict(sc_structure):
    opts = CategorizedInputOptions(sc_structure, task=Task.structure_opt,
                                   xc=Xc.pbe, vbm_cbm=[0, 0.5])
    assert opts.vise_log == ViseLog(
        version=__version__, task=Task.structure_opt, xc=Xc.pbe,
        input_options={"vbm_cbm": [0, 0.5],
                       "kpt_density": defaults.insulator_kpoint_density})


def test_defect_kpt_density(sc_structure):
    opts = CategorizedInputOptions(sc_structure, task=Task.defect, xc=Xc.pbe)
    actual = opts.structure_kpoints_options["kpt_density"]
    expected = defaults.defect_kpoint_density
    assert actual == expected




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import os
from argparse import Namespace
from copy import deepcopy
from pathlib import Path

import yaml
from mp_api.client import MPRester
from pymatgen.core import Structure
from pymatgen.io.vasp import Vasprun, Outcar
from tabulate import tabulate
from vise.analyzer.dielectric_function import DieleFuncData
from vise.analyzer.plot_band import BandMplPlotter
from vise.analyzer.plot_diele_func_data import DieleFuncMplPlotter, \
    DieleFuncPlotType
from vise.analyzer.plot_dos import DosPlotter
from vise.analyzer.vasp.band_edge_properties import VaspBandEdgeProperties
from vise.analyzer.vasp.dos_data import DosDataFromVasp
from vise.analyzer.vasp.make_diele_func import make_diele_func
from vise.analyzer.vasp.make_effective_mass import make_effective_mass
# from vise.analyzer.vasp.make_irreps import special_points_from_kpoints, \
#     make_irreps_from_wavecar
from vise.analyzer.vasp.plot_band import BandPlotInfoFromVasp
from vise.cli.main_tools import potcar_str2dict, list2dict
from vise.defaults import defaults
from vise.input_set.datasets.dataset_util import all_incar_flags
from vise.input_set.input_options import CategorizedInputOptions, \
    assignable_option_set
from vise.input_set.kpoints_mode import KpointsMode
from vise.input_set.prior_info import prior_info_from_calc_dir, PriorInfo
from vise.input_set.task import Task
from vise.input_set.vasp_input_files import VaspInputFiles
from vise.util.file_transfer import FileTransfers
from vise.util.logger import get_logger
from vise.util.structure_symmetrizer import StructureSymmetrizer

logger = get_logger(__name__)


def structure_info(args: Namespace) -> None:
    input_structure = Structure.from_file(args.poscar)
    symmetrizer = StructureSymmetrizer(input_structure,
                                       symprec=args.symprec,
                                       angle_tolerance=args.angle_tolerance)
    if args.show_primitive:
        primitive = symmetrizer.primitive
        if primitive != input_structure:
            print(primitive.to(fmt="poscar"))
        else:
            logger.info("Input structure is a primitive cell.")
    elif args.show_conventional:
        conventional = symmetrizer.conventional
        if conventional != input_structure:
            print(conventional.to(fmt="poscar"))
        else:
            logger.info("Input structure is a conventional cell.")
    else:
        la = input_structure.lattice
        print(symmetrizer)
        print(f"Volume: {input_structure.volume}")
        print("abc   : " + " ".join([f"{i:0.6f}".rjust(10) for i in la.abc]))
        print("angles: " + " ".join([f"{i:0.6f}".rjust(10) for i in la.angles]))


def get_most_stable_mp_id_from_formula(formula: str):
    # API key is parsed via .pmgrc.yaml
    with MPRester() as m:
        # Due to mp_decode=True by default, class objects are restored.
        candidates = m.query(criteria=formula,
                             properties=["task_id", "e_above_hull",
                                         "spacegroup.symbol", "band_gap"])

    sorted_candidates = sorted(candidates, key=lambda x: x["e_above_hull"])
    x = []
    for c in sorted_candidates:
        x.append([c["task_id"], round(c["e_above_hull"], 3),
                  c["spacegroup.symbol"], round(c["band_gap"], 3)])

    if not x:
        logger.info(f"Formula {formula} does not exist in the materials "
                    f"project.")
        return None
    print(tabulate(x, headers=["mp_id", "e_above_hull", "space group",
                               "band gap"]))
    return x[0][0]


def get_most_stable_mp_id_from_formula_w_new_mprester(formula: str):
    # API key is parsed via .pmgrc.yaml
    with MPRester() as m:
        logger.info("Note that you're using the newer MPRester.")
        candidates = m.summary.search(
            formula=formula,
            fields=["material_id", "energy_above_hull",
                    "symmetry", "band_gap"])
    sorted_candidates = sorted(candidates, key=lambda x: x.energy_above_hull)
    table_info = []
    for c in sorted_candidates:
        table_info.append([str(c.material_id), round(c.energy_above_hull, 3),
                  c.symmetry.symbol, round(c.band_gap, 3)])

    if not table_info:
        logger.info(f"Formula {formula} does not exist in the materials "
                    f"project.")
        return None
    headers = ["mp_id", "energy_above_hull", "space group", "band gap"]
    print(tabulate(table_info, headers=headers))
    return table_info[0][0]


def get_poscar_from_mp(args: Namespace) -> None:
    if args.mpid is None and args.formula is None:
        raise ValueError("Either mp-id or formula needs to be specified.")

    if args.mpid is None and args.formula:
        args.mpid \
            = get_most_stable_mp_id_from_formula_w_new_mprester(args.formula)
        if args.mpid is None:
            return

    s = MPRester().get_structure_by_material_id(args.mpid)
    s.to(fmt="poscar", filename="POSCAR")
    query = MPRester().summary.search(material_ids=[args.mpid],
                                      fields=["total_magnetization",
                                              "band_gap"])
    data = query[0]
    d = {"total_magnetization": data.total_magnetization,
         "band_gap": data.band_gap,
         "data_source": f"new MPRester {args.mpid}"}

    Path("prior_info.yaml").write_text(yaml.dump(d), None)


class VaspSet:
    def __init__(self, args: Namespace):
        self.args = args
        if args.prev_dir:
            self.args.prev_dir = args.prev_dir.absolute()
        if args.poscar:
            self.args.poscar = args.poscar.absolute()

        for _dir in [d.absolute() for d in args.dirs]:
            if _dir.is_file():
                logger.info(f"{_dir} is a file, so skipped.")
                continue

            os.chdir(_dir)
            logger.info(f"Creating VASP input set in {_dir}...")
            try:
                self._prior_info = PriorInfo.load_yaml()
            except FileNotFoundError:
                self._prior_info = PriorInfo()
            self.task = Task.cluster_opt if self._prior_info.is_cluster \
                else args.task

            options = CategorizedInputOptions(
                structure=self._structure(),
                task=self.task,
                xc=args.xc,
                kpt_density=args.kpt_density,
                overridden_potcar=self._overridden_potcar(),
                **self._option_kwargs())

            vif = VaspInputFiles(options, self._overridden_incar_settings())
            vif.create_input_files(Path.cwd())
            if hasattr(self, "_file_transfers"):
                self._file_transfers.transfer()
            vif.vise_log.to_yaml_file()

    def _structure(self):
        # avoid overlapping structure for e.g., phonon forces.
        if self.args.prev_dir and self.task.fine_to_inherit_structure_from_prev:
            return Structure.from_file(self.args.prev_dir / defaults.contcar)
        if self.args.poscar:
            return Structure.from_file(self.args.poscar)
        return Structure.from_file("POSCAR")

    def _overridden_incar_settings(self):
        result = deepcopy(defaults.user_incar_settings)
        result.update(self._prior_info.incar)

        if self.args.user_incar_settings:
            args = list2dict(self.args.user_incar_settings, all_incar_flags)
            result.update(args)

        return result

    def _overridden_potcar(self):
        result = {}
        if self.args.overridden_potcar:
            result.update(potcar_str2dict(self.args.overridden_potcar))
        return result

    def _option_kwargs(self):
        result = deepcopy(defaults.options)
        if self._prior_info:
            result.update(self._prior_info.input_options_kwargs)

        if self.args.prev_dir:
            pi = prior_info_from_calc_dir(prev_dir_path=self.args.prev_dir,
                                          vasprun=self.args.vasprun,
                                          outcar=self.args.outcar)
            result.update(pi.input_options_kwargs)

            self._file_transfers = FileTransfers(self._file_transfer(),
                                                 path=self.args.prev_dir)

        if self.args.options:
            args = list2dict(self.args.options, assignable_option_set)
            if "kpt_density" in args and self.args.kpt_density:
                logger.warning(
                    f"Because kpt_density is set at kpt_density argument and "
                    f"optional argument, kpt_density argument value "
                    f"{self.args.kpt_density} is used.")
                args.pop("kpt_density")
            result.update(args)

            if "set_spin_orbit" in args:
                logger.info("Spin-orbit coupling is switched on.")
                result["time_reversal"] = False

        if self.args.uniform_kpt_mode:
            result["kpt_mode"] = KpointsMode.uniform

        return result

    def _file_transfer(self):
        if not self.args.file_transfer_type:
            return {}
        else:
            result = {}
            for filename, transfer in zip(self.args.file_transfer_type[0::2],
                                          self.args.file_transfer_type[1::2]):
                result[filename] = transfer
            return result


def plot_band(args: Namespace):
    band_plot_info_from_vasp = BandPlotInfoFromVasp(
        vasprun=Vasprun(args.vasprun), kpoints_filename=args.kpoints_filename)
    plot_info = band_plot_info_from_vasp.make_band_plot_info()

    # if args.wavecar_filename:
    #     try:
    #         with open(args.kpoints_filename) as f:
    #             first_line = f.readline()
    #         sg_num = int(first_line.split()[-1])
    #         logger.info(f"Space group number is set to {sg_num} read from "
    #                     f"{args.kpoints_filename} header.")
    #     except ValueError:
    #         logger.info(f"Failed to get space group number from "
    #                     f"{args.kpoints_filename} header.")
    #         sg_num = None

        # special_points, kpt_indices = \
        #     special_points_from_kpoints(args.kpoints_filename)
        # irreps = make_irreps_from_wavecar(
        #     special_points, kpt_indices, sg_num, args.wavecar_filename,
        #     args.poscar)
        # plot_info.band_energy_infos["1"].irreps = irreps

    plot_info.to_json_file()

    # if args.plotly:
    #     plotter = BandDosPlotlyPlotter(band_plot_info=plot_info)
    #     plotter.fig.show()

#         bz_info = band_plot_info_from_vasp.make_bz_plot_info()
#         bz_plotter = BZPlotlyPlotter(bz_info)
#         fig = bz_plotter.create_figure(set_k123=False)
#         fig.show()
#         fig.write_image("brillouin_zone.pdf")
#         return

    plotter = BandMplPlotter(plot_info, energy_range=args.y_range)
    plotter.construct_plot()
    plotter.plt.savefig(args.filename, format="pdf")


def plot_dos(args: Namespace):
    vasprun = Vasprun(args.vasprun)
    outcar = Outcar(args.outcar)
    band_edge = VaspBandEdgeProperties(vasprun, outcar)

    if band_edge.band_gap:
        vertical_lines = [band_edge.vbm_info.energy, band_edge.cbm_info.energy]
    else:
        vertical_lines = [vasprun.efermi]

    if args.base_energy is None:
        base = vertical_lines[0]
    else:
        base = args.base_energy

    dos_data_from_vasp = \
        DosDataFromVasp(vasprun, vertical_lines, base, args.crop_first_value)
    dos_data = dos_data_from_vasp.make_dos_data()

    ylim_set = None
    if args.y_max_ranges:
        if dos_data.spin:
            ylim_set = [[-y_max, y_max] for y_max in args.y_max_ranges]
        else:
            ylim_set = [[0, y_max] for y_max in args.y_max_ranges]

    structure = vasprun.final_structure
    grouped_atom_indices = args.type.grouped_atom_indices(structure,
                                                          args.target)
    logger.info(f"Grouped atom indices: {grouped_atom_indices}")
    plot_data = dos_data.dos_plot_data(grouped_atom_indices,
                                       energy_range=args.x_range,
                                       dos_ranges=ylim_set,
                                       title=args.title)
    plot_data.to_json_file()
    plotter = DosPlotter(plot_data, args.legend)
    plotter.construct_plot()
    plotter.plt.savefig(args.filename, format="pdf")


def plot_diele_func(args: Namespace):
    if args.input_csv_name:
        diele_func_data = DieleFuncData.from_csv_file(args.input_csv_name)
    else:
        diele_func_data = make_diele_func(Vasprun(args.vasprun),
                                          Outcar(args.outcar),
                                          use_vasp_real=not args.calc_kk,
                                          ita=args.ita)
    diele_func_data.to_json_file()
    if args.to_csv:
        diele_func_data.to_csv_file()
    if args.y_range:
        if args.plot_type is DieleFuncPlotType.absorption_coeff:
            y_range = [10 ** args.y_range[0], 10 ** args.y_range[1]]
        else:
            y_range = args.y_range
    else:
        y_range = None

    diele_func_data.title = args.title
    plotter = DieleFuncMplPlotter(diele_func_data)
    plotter.construct_plot(directions=args.directions,
                           plot_type=args.plot_type, y_range=y_range)
    filename = args.filename or str(args.plot_type) + ".pdf"
    plotter.plt.savefig(filename, format="pdf")


def calc_effective_mass(args: Namespace):
    vasprun, outcar = Vasprun(args.vasprun), Outcar(args.outcar)
    band_edge_prop = VaspBandEdgeProperties(vasprun, outcar)
    try:
        vbm, cbm = band_edge_prop.vbm_cbm
    except TypeError:
        logger.warning("Band gap does not exist, so not suited for effective"
                       "mass calculation.")
        return
    effective_mass = make_effective_mass(vasprun,
                                         args.temperature,
                                         args.concentrations,
                                         vbm, cbm)
    effective_mass.to_json_file()
    print(effective_mass)


def band_edge_properties(args: Namespace):
    vasprun = Vasprun(args.vasprun)
    outcar = Outcar(args.outcar)
    print(VaspBandEdgeProperties(vasprun, outcar))



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from matplotlib.ticker import FuncFormatter


# need pos as the FuncFormatter arg requires 2 args.
# https://stackoverflow.com/questions/29139552/pyplot-remove-the-digits-of-zero-start-from-0-not-0-00

def my_formatter(tick_value, pos):
    """convert  0.0 to 0 in the plot.

    Examples:
        ax.xaxis.set_major_formatter(formatter)
        ax.yaxis.set_major_formatter(formatter)
    """
    if isinstance(tick_value, float):
        rounded_value = round(tick_value, ndigits=10)
        if rounded_value.is_integer():
            return int(rounded_value)
        else:
            return rounded_value
    else:
        return str(tick_value)


float_to_int_formatter = FuncFormatter(my_formatter)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from typing import Optional, List, Union, Iterable

from pymatgen.core.periodic_table import Element

from vise.util.logger import get_logger
from vise.util.str_related_tools import is_str_int, is_str_digit, str2bool

logger = get_logger(__name__)


def potcar_str2dict(potcar_list: Union[str, List[str], None]) -> dict:
    """Sanitize potcar names to dict.

    If potcar_dict is None, {} is returned.
    Examples
        ["Mg_pv", O_h"] -> {"Mg": "Mg_pv", "O": "O_h"}
        "Mg_pv" -> {"Mg": "Mg_pv"}
        None -> {}

    Args:
         potcar_list (str/list/None):
            List of Potcar names or a name, including element names before the
            hyphen.

    Returns:
         Dictionary with element names as keys and potcar names as values.
    """
    if potcar_list is None:
        return {}
    elif isinstance(potcar_list, str):
        potcar_list = [potcar_list]

    d = {}
    for p in potcar_list:
        element = p.split("_")[0]
        if element in d:
            raise ValueError("Multiple POTCAR files for an element are not "
                             "supported yet.")
        try:
            Element(element)
        except ValueError:
            raise ValueError(f"First part of POTCAR file name {p} is not an "
                             f"element supported yet.")

        d[element] = p
    return d


def list2dict(flattened_list: Optional[list], key_candidates: Iterable) -> dict:
    """Sanitize a list to a dict with keys in the flags

    If a string in list does not exist in key_candidates, raise ValueError.

    Examples:
        flattened_list = ["ENCUT", "500", "MAGMOM", "4", "4", "LWAVE", "F"]
        key_candidates = ["ENCUT", "MAGMOM", "LWAVE", ...]
        list2dict(flattened_list, key_candidates) =
                            {"ENCUT": 500, "MAGMOM": [4, 4], "LWAVE": False}

        flattened_list = ["ENCUT", "500", "MAGMAM", "4", "4"]
        list2dict(flattened_list, key_candidates) raises ValueError

    Args:
        flattened_list (list):
            Input list.
        key_candidates (list):
            List of key candidates, e.g., INCAR flags.

    Returns:
        Sanitized dict.
    """
    flattened_list = flattened_list or []

    d = {}
    key = None  # key is None at the beginning or after inserting a value.
    value_list = []

    def insert():
        if not value_list:
            raise ValueError(f"Invalid input: {flattened_list}.")
        d[key] = value_list[0] if len(value_list) == 1 else value_list

    for string in flattened_list:
        if key is None and string not in key_candidates:
            raise ValueError(f"Keys are invalid: {flattened_list}.")
        elif string in key_candidates:
            if key:
                insert()
                key = None
                value_list = []
            key = string
        else:
            # We first need to check numbers as 0 and 1 are parse as False and
            # True with str2bool.
            if is_str_int(string):
                value = int(string)
            else:
                try:
                    value = str2bool(string)
                except ValueError:
                    if is_str_digit(string):
                        value = float(string)
                    else:
                        value = string

            value_list.append(value)
    else:
        if key:
            insert()

    return d



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
from pymatgen.core import Lattice
from pymatgen.core.structure import Structure

from vise.input_set.kpoints_mode import KpointsMode
from vise.input_set.structure_kpoints_generator import \
    StructureKpointsGenerator
from vise.input_set.task import Task
from vise.util.structure_symmetrizer import StructureSymmetrizer


def test_constructor_num_kpt_multiplication_factor(sc_structure):
    generator = StructureKpointsGenerator(sc_structure,
                                          task=Task.dielectric_function,
                                          kpt_density=1.0)
    assert generator._num_kpt_factor == 3


def test_constructor_only_even_num_kpts(sc_structure):
    generator = StructureKpointsGenerator(sc_structure,
                                          only_even_num_kpts=True,
                                          task=Task.cluster_opt,
                                          kpt_density=1.0)
    assert generator._only_even_num_kpts is False


def test_constructor_kpt_mode(sc_structure):
    generator = StructureKpointsGenerator(sc_structure,
                                          task=Task.cluster_opt,
                                          kpt_density=1.0)
    assert generator._kpt_mode is KpointsMode.uniform


def test_constructor_kpt_shift_structure_opt(sc_structure):
    generator = StructureKpointsGenerator(sc_structure,
                                          task=Task.structure_opt,
                                          kpt_density=1.0)
    assert generator._gamma_centered is None


def test_constructor_kpt_shift_dos(sc_structure):
    generator = StructureKpointsGenerator(sc_structure,
                                          task=Task.dos, kpt_density=1.0)
    assert generator._gamma_centered is True


def test_constructor_symmetrizer(sc_structure, mocker):
    mock = mocker.patch(
        "vise.input_set.structure_kpoints_generator.StructureSymmetrizer")
    StructureKpointsGenerator(sc_structure,
                              task=Task.dos,
                              symprec=10.0,
                              angle_tolerance=20.0,
                              band_ref_dist=30.0,
                              time_reversal=False,
                              kpt_density=1.0)
    mock.assert_called_once_with(
        structure=sc_structure, symprec=10.0, angle_tolerance=20.0,
        band_mesh_distance=30.0, time_reversal=False)

# Structure related unittests.


def test_mc_structure_primitive(mc_structure):
    generator = StructureKpointsGenerator(mc_structure, task=Task.structure_opt,
                                          kpt_density=1.0)
    generator.generate_input()
    actual = generator.structure.lattice.matrix
    expected = [[ 6.,-3., 0.],
                [ 6., 3., 0.],
                [-4., 0., 7.]]
    np.testing.assert_array_almost_equal(actual, expected)


def test_mc_structure_band(mc_structure):
    generator = StructureKpointsGenerator(mc_structure, task=Task.band,
                                          kpt_density=1.0)
    generator.generate_input()
    actual = generator.structure.lattice.matrix
    expected = [[ 6., 3., 0.],
                [-6., 3., 0.],
                [-4., 0., 7.]]
    np.testing.assert_array_almost_equal(actual, expected)


def test_mc_structure_dos_uniform(mc_structure):
    generator = StructureKpointsGenerator(mc_structure, task=Task.dos,
                                          kpt_mode=KpointsMode.uniform,
                                          kpt_density=1.0)
    generator.generate_input()
    actual = generator.structure.lattice.matrix
    expected = [[ 6., 3., 0.],
                [ 6.,-3., 0.],
                [-4., 0., 7.]]
    np.testing.assert_array_almost_equal(actual, expected)


# KPOINTS related unittests.

def test_manual_kpts():
    lattice =[[10.0, 0.0, 0.0],
              [ 0.0, 8.0, 0.0],
              [-2.0, 0.0, 8.0]]
    coords = [[0.0, 0.0, 0.0]]
    # lattice angles are (90.0, 104.0362434679265, 90.0)
    # so only kpoint shift along b-direction must be 0.5.
    structure = Structure(lattice=lattice, species=["H"], coords=coords)
    generator = StructureKpointsGenerator(structure,
                                          kpt_density=2.5,
                                          task=Task.structure_opt,
                                          kpt_mode=KpointsMode.uniform)
    generator.generate_input()
    assert generator.kpoints.kpts[0] == (2, 2, 2)
    assert generator.kpoints.kpts_shift == (0.0, 0.5, 0.0)
    generator = StructureKpointsGenerator(structure,
                                          task=Task.structure_opt,
                                          kpt_mode=KpointsMode.uniform,
                                          gamma_centered=True,
                                          kpt_density=1.0)
    generator.generate_input()
    assert generator.kpoints.kpts_shift == (0.0, 0.0, 0.0)


def test_only_even_num_kpts(sc_structure):
    generator = StructureKpointsGenerator(sc_structure,
                                          task=Task.structure_opt,
                                          kpt_density=2.)
    generator.generate_input()
    num_kpt_list = generator.kpoints.kpts[0]
    assert num_kpt_list == (13, 13, 13)

    generator = StructureKpointsGenerator(sc_structure,
                                          task=Task.structure_opt,
                                          kpt_density=2.,
                                          only_even_num_kpts=True)
    generator.generate_input()
    num_kpt_list = generator.kpoints.kpts[0]
    assert num_kpt_list == (14, 14, 14)


def test_kpt_factor(sc_structure):
    generator = StructureKpointsGenerator(sc_structure,
                                          task=Task.structure_opt,
                                          kpt_density=2.,
                                          num_kpt_factor=2)
    generator.generate_input()
    num_kpt_list = generator.kpoints.kpts[0]
    assert num_kpt_list == (26, 26, 26)


def test_cluster_opt_kpoints(sc_structure):
    generator = StructureKpointsGenerator(sc_structure,
                                          task=Task.cluster_opt,
                                          kpt_density=2.)
    generator.generate_input()
    num_kpt_list = generator.kpoints.kpts[0]
    assert num_kpt_list == (1, 1, 1)
    kpt_shift = generator.kpoints.kpts_shift
    assert kpt_shift == (0.0, 0.0, 0.0)


def test_band_path(sc_structure):
    generator = StructureKpointsGenerator(sc_structure,
                                          task=Task.band,
                                          kpt_density=0.5,
                                          band_ref_dist=0.5)
    # reciprocal length = 2 * pi / 1 = 6.28
    # so num_kpt_list = [12, 12, 12] and kpt_shift = [0.5, 0.5, 0.5] are set.
    generator.generate_input()
    assert generator.kpoints.num_kpts == 44
    assert generator.kpoints.kpts[0] == (0.125, 0.125, 0.125)


def test_oi_ti_bravais():
    structure = Structure(Lattice.orthorhombic(20, 8, 6),
                          species=["H"]*2,
                          coords=[[0.0]*3, [0.5]*3])
    generator = StructureKpointsGenerator(structure,
                                          task=Task.structure_opt,
                                          kpt_density=5)
    generator.generate_input()

    primitive_structure = StructureSymmetrizer(structure).primitive

    expected_generator = StructureKpointsGenerator(primitive_structure,
                                                   task=Task.structure_opt,
                                                   kpt_density=5)
    expected_generator.generate_input()

    assert generator.num_kpts == expected_generator.num_kpts
    assert generator.kpoints.kpts_shift == expected_generator.kpoints.kpts_shift


def test_hexagonal():
    hexagonal_lattice =[[10.0, 0.0,      0.0],
                        [-5.0, 8.660254, 0.0],
                        [ 0.0, 0.0,     10.0]]
    coords = [[0.0, 0.0, 0.0]]
    # recipro_lat_abc = (0.85, 1.09, 1.31)
    structure = Structure(lattice=hexagonal_lattice,
                          species=["H"], coords=coords)
    generator = StructureKpointsGenerator(structure,
                                          task=Task.structure_opt,
                                          kpt_density=5)
    generator.generate_input()
    assert generator.kpoints.kpts_shift == (0.0, 0.0, 0.5)


def test_conventional_input(tmpdir):
    structure = Structure(Lattice.cubic(3.183372), species=["H"] * 2,
                          coords=[[0] * 3, [0.5] * 3])
    generator = StructureKpointsGenerator(initial_structure=structure,
                                          task=Task.structure_opt,
                                          kpt_density=2.5,
                                          num_kpt_factor=2)
    generator.generate_input()

    expected = Structure([[-1.591686, 1.591686, 1.591686],
                          [1.591686, -1.591686, 1.591686],
                          [1.591686, 1.591686, -1.591686]],
                         species=["H"], coords=[[0]*3])
    assert generator.structure == expected
    assert generator.num_kpts == 100


def test_defect_2d():
    structure = Structure.from_str("""Mo2 S4
    1.000000000000000
     3   0.                0.
    -1.5 2.598076211375221 0.
     0.  0.               10.
   Mo   S
     2     4
Direct
  0.3333333332999970  0.6666666667000030  0.25
  0.6666666667000030  0.3333333332999970  0.75
  0.6666666667000030  0.3333333332999970  0.375
  0.3333333332999970  0.6666666667000030  0.875
  0.6666666667000030  0.3333333332999970  0.125
  0.3333333332999970  0.6666666667000030  0.625""", fmt="poscar")
    generator = StructureKpointsGenerator(structure,
                                          task=Task.defect_2d,
                                          kpt_density=5.)
    generator.generate_input()
    num_kpt_list = generator.kpoints.kpts[0]
    assert num_kpt_list == (13, 13, 1)
    assert generator.kpoints.kpts_shift == (0.0, 0.0, 0.0)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import warnings
from abc import ABC

from monty.design_patterns import singleton
from pathlib import Path
from vise.input_set.datasets.potcar_set import PotcarSet
from vise.input_set.task import Task
from vise.input_set.xc import Xc
from vise.user_settings import UserSettings
from vise.util.logger import get_logger

logger = get_logger(__name__)


class DefaultsBase(ABC):
    def set_user_settings(self, yaml_filename):
        user_settings = UserSettings(yaml_filename=yaml_filename)
        self.yaml_files = user_settings.yaml_files_from_root_dir
        self.user_settings = user_settings.user_settings

        for k, v in self.user_settings.items():
            if hasattr(self, k):
                self.__setattr__("_" + k, v)
            else:
                logger.warning(f"{k} key in {yaml_filename} is not valid.")


@singleton
class Defaults(DefaultsBase):
    def __init__(self):
        self._symmetry_length_tolerance = 0.01
        self._symmetry_angle_tolerance = 5.0
        self._dos_step_size = 0.01
        self._kpoint_density = 5.0
        self._insulator_kpoint_density = 2.5
        self._defect_kpoint_density = 1.8
        self._band_mesh_distance = 0.025
        self._str_opt_encut_factor = 1.3
        self._band_gap_criterion = 0.2  # in eV
        self._integer_criterion = 0.1
        self._default_num_cores = 4
        self._unused_core_ratio_threshold = 0.25
        self._task = str(Task.structure_opt)
        self._xc = str(Xc.pbe)
        self._options = {}
        self._user_incar_settings = {}
        self._u_parameter_set_yaml_file = Path(__file__).parent / "input_set" / "datasets" / "u_parameter_set.yaml"
        self._outcar = "OUTCAR"
        self._contcar = "CONTCAR"
        self._vasprun = "vasprun.xml"
        self._procar = "PROCAR"
        self._overridden_potcar = []
        self._potcar_set = str(PotcarSet.normal)
        self._lda_potcar = "LDA"
        self._gga_potcar = "PBE_54"
        self._suppress_user_warning = True

        self.set_user_settings(yaml_filename="vise.yaml")

        if self._suppress_user_warning:
            warnings.simplefilter('ignore', UserWarning)

    @property
    def symmetry_length_tolerance(self):
        return self._symmetry_length_tolerance

    @property
    def symmetry_angle_tolerance(self):
        return self._symmetry_angle_tolerance

    @property
    def dos_step_size(self):
        return self._dos_step_size

    @property
    def kpoint_density(self):
        return self._kpoint_density

    @property
    def insulator_kpoint_density(self):
        return self._insulator_kpoint_density

    @property
    def defect_kpoint_density(self):
        return self._defect_kpoint_density

    @property
    def band_mesh_distance(self):
        return self._band_mesh_distance

    @property
    def str_opt_encut_factor(self):
        return self._str_opt_encut_factor

    @property
    def band_gap_criterion(self):
        return self._band_gap_criterion

    @property
    def integer_criterion(self):
        return self._integer_criterion

    @property
    def default_num_cores(self):
        return self._default_num_cores

    @property
    def unused_core_ratio_threshold(self):
        return self._unused_core_ratio_threshold

    @property
    def task(self):
        return Task(self._task)

    @property
    def xc(self):
        return Xc(self._xc)

    @property
    def options(self):
        return self._options

    @property
    def user_incar_settings(self):
        return self._user_incar_settings

    @property
    def u_parameter_set_yaml_file(self) -> Path:
        return self._u_parameter_set_yaml_file

    @property
    def outcar(self):
        return Path(self._outcar)

    @property
    def contcar(self):
        return Path(self._contcar)

    @property
    def vasprun(self):
        return Path(self._vasprun)

    @property
    def procar(self):
        return Path(self._procar)

    @property
    def overridden_potcar(self):
        return self._overridden_potcar

    @property
    def potcar_set(self):
        return PotcarSet(self._potcar_set)

    @property
    def lda_potcar(self):
        return self._lda_potcar

    @property
    def gga_potcar(self):
        return self._gga_potcar


defaults = Defaults()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from collections import defaultdict
from copy import copy, deepcopy
from dataclasses import dataclass, fields
from functools import reduce
from typing import Dict, Optional, List

import numpy as np
import pandas as pd
from monty.json import MSONable
from vise.util.logger import get_logger
from vise.util.mix_in import ToJsonFileMixIn, ToCsvFileMixIn

logger = get_logger(__name__)


@dataclass
class PDos(MSONable, ToJsonFileMixIn):
    s:   np.ndarray  # [by spin][by energy]
    px:  np.ndarray
    py:  np.ndarray
    pz:  np.ndarray
    dxy: np.ndarray
    dyz: np.ndarray
    dxz: np.ndarray
    dx2: np.ndarray
    dz2: np.ndarray
    f_3: Optional[np.ndarray] = None
    f_2: Optional[np.ndarray] = None
    f_1: Optional[np.ndarray] = None
    f0:  Optional[np.ndarray] = None
    f1:  Optional[np.ndarray] = None
    f2:  Optional[np.ndarray] = None
    f3:  Optional[np.ndarray] = None

    @property
    def p(self):
        return self.px + self.py + self.pz

    @property
    def d(self):
        return self.dxy + self.dyz + self.dxz + self.dx2 + self.dz2

    @classmethod
    def from_dict(cls, d):
        if "px" in d:
            return super().from_dict(d)

        logger.warning("Because the PDOS are not decomposed to each orbital,"
                       "e.g., px, py, pz, they are equally divided.")
        x = dict(s=d["s"], px=d["p"] / 3, py=d["p"] / 3, pz=d["p"] / 3,
                 dxy=d["d"] / 5, dyz=d["d"] / 5, dxz=d["d"] / 5, dx2=d["d"] / 5,
                 dz2=d["d"] / 5)
        if "f" in d:
            x.update({
                "f_3": d["f"] / 7,
                "f_2": d["f"] / 7,
                "f_1": d["f"] / 7,
                "f0": d["f"] / 7,
                "f1": d["f"] / 7,
                "f2": d["f"] / 7,
                "f3": d["f"] / 7})
        return cls(**x)

    @property
    def f(self):
        try:
            return (self.f_3 + self.f_2 + self.f_1 +
                    self.f0 + self.f1 + self.f2 + self.f3)
        except TypeError:
            return None

    def __add__(self, other: "PDos"):
        args = {}
        for k, v in self.__dict__.items():
            if v is None:
                args[k] = None
            else:
                args[k] = v + getattr(other, k)
        return PDos(**args)

    def __eq__(self, other: "PDos"):
        for k, v in self.__dict__.items():
            if (v != getattr(other, k)).any():
                return False
        return True


@dataclass
class DosData(MSONable):
    energies: List[float]
    total: np.ndarray
    pdos: List[PDos]
    vertical_lines: List[float]
    base_energy: Optional[float] = 0.0

    def __post_init__(self):
        if isinstance(self.total, list):
            self.total = np.array(self.total)

    @property
    def spin(self):
        return False if len(self.total) == 1 else True

    def dos_plot_data(self,
                      grouped_atom_indices: Dict[str, List[int]],
                      energy_range: Optional[List[float]] = None,
                      dos_ranges: Optional[List[List[float]]] = None,
                      title: Optional[str] = None) -> "DosPlotData":
        """
        Args:
            grouped_atom_indices:
                key: name of the grouped
            energy_range:
            dos_ranges:
            title:
        Returns:

        """

        if dos_ranges is not None:
            print(len(grouped_atom_indices))
            print(len(dos_ranges))
            try:
                # total+pdos
                assert len(grouped_atom_indices) + 1 == len(dos_ranges)
            except AssertionError:
                print(f"Number of ranges for DOS plots "
                      f"{len(grouped_atom_indices) + 1} are not the same as " 
                      f"that of actual plot: {len(dos_ranges)}."
                      f"The caveat is the plot includes the total DOS, so"
                      f"if there are two PDOS, one needs to set three DOS"
                      f"ranges.")
                raise

        # Total dos does not have spin decomposition.
        doses = [[DosBySpinEnergy("", self.total.tolist())]]
        names = ["total"]

        for name, atom_indices_by_group in grouped_atom_indices.items():
            pdos_list = [self.pdos[idx] for idx in atom_indices_by_group]
            pdos = reduce(lambda x, y: x + y, pdos_list)
            pdos_by_ax = [DosBySpinEnergy("s", pdos.s.tolist()),
                          DosBySpinEnergy("p", pdos.p.tolist()),
                          DosBySpinEnergy("d", pdos.d.tolist())]
            if pdos.f is not None:
                pdos_by_ax.append(DosBySpinEnergy("f", pdos.f.tolist()))

            doses.append(pdos_by_ax)
            names.append(name)

        energy_range = energy_range or [-5, 10]
        abs_xlim = [x + self.base_energy for x in energy_range]

        if dos_ranges is None:
            dos_ranges = default_dos_ranges(abs_xlim, doses, self.energies,
                                            self.spin)

        energies = [e - self.base_energy for e in self.energies]
        shifted_energy_lines = [e - self.base_energy
                                for e in self.vertical_lines]

        return DosPlotData(energies, doses, names, energy_range, dos_ranges,
                           shifted_energy_lines, title)


def default_dos_ranges(energy_range, doses, energies, spin_polarized,
                       multi: float = 1.1, round_digit: int = 2):
    mask = np.ma.masked_outside(energies, energy_range[0], energy_range[1]).mask
    max_dos_by_ax = []
    for dos_by_ax in doses:
        max_dos_by_ax.append(np.max([dos_by_spin.max_dos(mask)
                                     for dos_by_spin in dos_by_ax]))

    total_dos_max = max_dos_by_ax[0]
    plot_maxes = [total_dos_max]
    if len(max_dos_by_ax) > 1:
        pdos_max = max(max_dos_by_ax[1:])
        plot_maxes.extend([pdos_max] * (len(doses) - 1))

    max_y_ranges = [round(i * multi, round_digit) for i in plot_maxes]

    if spin_polarized:
        return [[-y, y] for y in max_y_ranges]
    return [[0, y] for y in max_y_ranges]


@dataclass
class DosBySpinEnergy(MSONable):
    name: str  # e.g., "s", "p", ..., in case of total, name=""
    dos: List[List[float]]  # [by spin][by energy]

    def max_dos(self, mask: List[bool] = None):
        return max([np.max(np.ma.masked_array(d, mask).compressed())
                    for d in self.dos])


@dataclass
class DosPlotData(MSONable, ToJsonFileMixIn, ToCsvFileMixIn):
    relative_energies: List[float]
    doses: List[List[DosBySpinEnergy]]  # [by ax][by orbital, i.e., s, p, d, f]
    names: List[str]  # For each ax, e.g., ["total", "H"]
    energy_range: List[float]  # e.g., [-5.0, 6.0]
    dos_ranges: List[List[float]]  # e.g., [[0.0, 6.0], [-1.0, 5.0], ...]
    energy_lines: List[float]  # e.g., VBM and CBM or Fermi level
    title: str = None

    @property
    def to_dataframe(self) -> pd.DataFrame:
        columns = ["energy(eV)"]
        inv_data = [[e for e in self.relative_energies]]

        for name, dos_by_ax in zip(self.names, self.doses):
            for dos_by_spin_energy in dos_by_ax:
                for spin, dos_by_spin in zip(["up", "down"],
                                             dos_by_spin_energy.dos):

                    columns.append(f"{name}_{dos_by_spin_energy.name}_{spin}")
                    inv_data.append(dos_by_spin)

        columns.append("energy_lines")
        inv_data.append(self.energy_lines)

        df = pd.DataFrame(inv_data).T
        df.columns = columns
        return df

    @classmethod
    def from_dataframe(cls, df):
        sanitized_data = defaultdict(lambda: defaultdict(list))
        for full_name in df.columns[1:-1]:
            name, orbital, spin = full_name.split("_")
            sanitized_data[name][orbital].append(list(df[full_name]))

        doses, names = [], []
        for k, v in sanitized_data.items():
            names.append(k)
            doses.append([{"name": k2, "dos": v2} for k2, v2 in v.items()])

        d = {"relative_energies": df["energy(eV)"].tolist(),
             "doses": doses,
             "names": names,
             "energy_lines": [i for i in df.energy_lines if i]}
        return cls.from_dict(d)

    @classmethod
    def from_dict(cls, d):
        for k in copy(d):
            if k[0] == "@":
                d.pop(k)
        # For backward compatibility
        if "xlim" in d:
            d["energy_range"] = d.pop("xlim")
        if "ylim_set" in d:
            d["dos_ranges"] = d.pop("ylim_set")
        if "vertical_lines" in d:
            d["energy_lines"] = d.pop("vertical_lines")
        for i, x in enumerate(d["doses"]):
            for j, y in enumerate(x):
                d["doses"][i][j] = DosBySpinEnergy.from_dict(y)

        if "energy_range" not in d:
            d["energy_range"] = [-5, 10]
        if "dos_ranges" not in d:
            spin_polarized = len(d["doses"][0][0].dos) == 2
            d["dos_ranges"] = default_dos_ranges(
                d["energy_range"], d["doses"], d["relative_energies"],
                spin_polarized, multi=1.1, round_digit=2)

        return cls(**d)


def scissor_energy(dos_plot_data: DosPlotData,
                   energy_shift: float) -> DosPlotData:
    result = deepcopy(dos_plot_data)
    gap_middle = np.mean(dos_plot_data.energy_lines)
    energy_over_gap_middle = np.array(result.relative_energies) >= gap_middle
    shifted_energy_start_idx = int(np.argmax(energy_over_gap_middle))
    for i in range(shifted_energy_start_idx, len(result.relative_energies)):
        result.relative_energies[i] += energy_shift

    result.energy_lines[1] += energy_shift
    return result



# -*- coding: utf-8 -*-
#  Copyright (c) 2022. Distributed under the terms of the MIT License.
from pathlib import Path

import pytest
from vise.analyzer.plot_band import Irrep
# from vise.analyzer.vasp.make_irreps import make_irreps_from_wavecar, find_irrep, \
#     ViseNoIrrepError, special_points_from_kpoints
from vise.tests.helpers.assertion import assert_dataclass_almost_equal

#
# def test_special_points_from_kpoints(test_data_files: Path):
#     kpoints_file = str(test_data_files / "H_band_KPOINTS")
#     actual = special_points_from_kpoints(kpoints_file)
#     assert actual == (["GM", "H", "N", "P"], [9, 14, 17, 24])


# def test_make_irreps_with_actual_vasp_files(test_data_files: Path):
#     wavecar_file = str(test_data_files / "H_band_WAVECAR")
#     poscar_file = str(test_data_files / "H_band_POSCAR")
#     # K does not exist in irreptables.
#     actual = make_irreps_from_wavecar(special_point_symbols=["GM", "K", "H"],
#                                       kpt_indices=[9, 15, 14],
#                                       wavecar_filename=wavecar_file,
#                                       poscar_filename=poscar_file,
#                                       )
#     assert actual.sg_num == 229
#
#     assert_dataclass_almost_equal(actual.irreps["H"],
#                                   Irrep([0.5, -0.5, 0.5],
#                                         ["H1+", "H4-", "H3+"],
#                                         [-5.93323, 2.71352, 3.80069],
#                                         [1, 3, 2]), digit=3)
#     assert_dataclass_almost_equal(actual.irreps["Γ"],
#                                   Irrep([0.0, 0.0, 0.0],
#                                         ["Γ1+", "Γ1+", "Γ4-", "Unknown"],
#                                         [-6.5536, 0.9245, 6.7080, 8.6659],
#                                         [1, 1, 3, 1]), digit=2)

#
# def test_find_irrep():
#     d = {"A": (0.991, 0.0), "B": (0.009, 0.0)}
#     assert find_irrep(d) == "A"
#     with pytest.raises(ViseNoIrrepError):
#         find_irrep(d, threshold=0.992)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import os
import warnings
from pathlib import Path
from xml.etree.ElementTree import ParseError

from pymatgen.core import Element
from pymatgen.io.vasp import Vasprun, Outcar
from pymatgen.io.vasp.inputs import UnknownPotcarWarning
from vise.atom_energies.make_atom_vasp_set import is_target_element
from vise.util.logger import get_logger


logger = get_logger(__name__)

warnings.simplefilter('ignore', UnknownPotcarWarning)


def make_energy_yaml():
    dirs = [f.name for f in os.scandir(".") if f.is_dir()]

    for e in list(Element):
        e = str(e)
        if e not in dirs or is_target_element(e) is False:
            continue

        try:
            v = Vasprun(Path(e) / "vasprun.xml")

            if v.converged_electronic is False:
                logger.warning(f"Calculation for {e} is not converged.")
                continue
        except ParseError:
            logger.warning(f"Parsing vasprun.xml for {e} failed.")
            continue

        outcar = Outcar(Path(e) / "OUTCAR")
        print(f"{e + ':':<3} {outcar.final_energy:11.8f}")


if __name__ == '__main__':
    make_energy_yaml()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
import pytest
from pymatgen.electronic_structure.boltztrap import BoltztrapError

from pymatgen.io.vasp import Vasprun, Outcar
from vise.analyzer.vasp.band_edge_properties import VaspBandEdgeProperties

try:
    from pymatgen.electronic_structure.boltztrap2 import VasprunBSLoader, \
        BztInterpolator, BztTransportProperties
    BOLTZTRAP2_NOT_PRESENT = False
except BoltztrapError:
    BOLTZTRAP2_NOT_PRESENT = True


@pytest.mark.skipif(BOLTZTRAP2_NOT_PRESENT, reason="boltztrap2 is not installed.")
def test_make_effective_mass(test_data_files):
#def test_make_effective_mass(mocker, test_data_files):
    from vise.analyzer.effective_mass import EffectiveMass
    from vise.analyzer.vasp.make_effective_mass import make_effective_mass
    v = Vasprun(test_data_files / "MgSe_absorption_vasprun.xml")
    # mock_vl = mocker.patch("vise.analyzer.vasp.make_effective_mass.VasprunBSLoader")
    # mock_bi = mocker.patch("vise.analyzer.vasp.make_effective_mass.BztInterpolator")
    # mock_btp = mocker.patch("vise.analyzer.vasp.make_effective_mass.BztTransportProperties")

    p = [[[ 2.27995989e+00, -7.16015256e-17, -1.92884504e-16],
          [-7.16015256e-17,  2.27995989e+00, -1.17150055e-16],
          [-1.92884504e-16, -1.17150055e-16,  2.27995989e+00]]]
    n = [[[ 5.49195397e-01, -1.09903151e-17, -3.07622696e-17],
          [-1.09903151e-17,  5.49195397e-01, -5.50277544e-17],
          [-3.07622696e-17, -5.50277544e-17,  5.49195397e-01]]]
    # mock_btp.return_value.Effective_mass_doping = {"p": np.array([p]),
    #                                                "n": np.array([n])}
    # assume the situation where the efermi locates lower than CBM.
    v.efermi = 0.55
    actual = make_effective_mass(vasprun=v, temp=300, concentrations=[1e18],
                                 vbm=0.5614, cbm=3.0904)

    # mock_bi.assert_called_with(mock_vl.return_value, energy_range=3.0)
    # mock_btp.assert_called_with(mock_bi.return_value, temp_r=np.array([300]))
    # mock_btp.return_value.compute_properties_doping.assert_called_with([1e18])

    expected = EffectiveMass(p=p, n=n, temperature=300, concentrations=[1e+18])
    np.testing.assert_array_almost_equal(actual.p, expected.p)
    np.testing.assert_array_almost_equal(actual.n, expected.n)
    np.testing.assert_array_almost_equal(actual.concentrations,
                                         expected.concentrations)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from monty.json import MSONable

from vise.util.enum import ExtendedEnum


class Xc(MSONable, ExtendedEnum):
    pbe = "pbe"
    pbesol = "pbesol"
    lda = "lda"
    scan = "scan"
    pbe0 = "pbe0"
    hse = "hse"

    @classmethod
    def from_string(cls, s: str):
        if s == "perdew-zunger81":
            return cls.lda
        return super().from_string(s)

    @property
    def is_lda_or_gga(self):
        return self in (self.pbe, self.pbesol, self.lda)

    @property
    def is_hybrid_functional(self):
        return self in (self.pbe0, self.hse)

    @property
    def is_local_or_semilocal(self):
        return self.is_lda_or_gga or self is self.scan

    @property
    def is_nonlocal(self):
        return self.is_hybrid_functional



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from enum import unique, Enum


@unique
class ExtendedEnum(Enum):

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s: str):
        for m in cls:
            if m.name == s:
                return m
        raise AttributeError(f"{str(s)} is not proper member for {cls}.\n",
                             f"Supported names:\n {cls.names_string()}")

    @classmethod
    def names_string(cls):
        return ", ".join([e.name for e in cls])

    @classmethod
    def name_list(cls):
        return [e for e in cls]

    @classmethod
    def name_str_list(cls):
        return [str(e) for e in cls]

    @classmethod
    def values(cls):
        return [e.value for e in cls]



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import numpy as np

from vise.util.enum import ExtendedEnum


class Centering(ExtendedEnum):
    P = "P"
    A = "A"
    C = "C"
    R = "R"
    I = "I"
    F = "F"

    @property
    def conv_to_primitive(self):
        """ Note that the matrices are transposed from those written in spglib
        web sites due to the difference of matrix convention.

        See unittests for sure.

        https://spglib.github.io/spglib/definition.html
        """
        if self is self.P:
            return np.eye(3)
        elif self is self.A:
            return np.array([[1,    0,    0],
                             [0,  1/2,  1/2],
                             [0, -1/2,  1/2]])
        elif self is self.C:
            return np.array([[1/2, -1/2, 0],
                             [1/2,  1/2, 0],
                             [  0,    0, 1]])
        elif self is self.R:
            return np.array([[ 2/3,  1/3,  1/3],
                             [-1/3,  1/3,  1/3],
                             [-1/3, -2/3,  1/3]])
        elif self is self.I:
            return np.array([[-1/2,  1/2,  1/2],
                             [ 1/2, -1/2,  1/2],
                             [ 1/2,  1/2, -1/2]])
        elif self is self.F:
            return np.array([[  0,  1/2, 1/2],
                             [1/2,    0, 1/2],
                             [1/2,  1/2,   0]])
        else:
            raise NotImplementedError(f"Define matrix for {self}")

    @property
    def primitive_to_conv(self):
        return np.linalg.inv(self.conv_to_primitive)

    @property
    def conv_multiplicity(self):
        return int(round(1 / np.linalg.det(self.conv_to_primitive)))

    # return matrix




# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pathlib import Path

from monty.serialization import loadfn
from vise.util.enum import ExtendedEnum
from vise.util.logger import get_logger

path = Path(__file__).parent

mp_energies = loadfn(path / "mp_atom_energy.yaml")





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from typing import List, Optional

import numpy as np
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp import Vasprun

from vise.analyzer.dos_data import DosData
from vise.analyzer.dos_data import PDos
from vise.util.logger import get_logger
logger = get_logger(__name__)


class DosDataFromVasp:

    def __init__(self, vasprun: Vasprun,
                 vertical_lines: Optional[List[float]] = None,
                 base_energy: float = 0.0,
                 crop_first_value=False,
                 energy_window: List[float] = None):
        self.complete_dos = vasprun.complete_dos
        self.vertical_lines = vertical_lines or []
        self.base_energy = base_energy
        self.energy_window = energy_window
        self.min_energy_idx = 1 if crop_first_value else 0
        self.max_energy_idx = len(self.complete_dos.energies)

    def make_dos_data(self):
        energies = self.complete_dos.energies.tolist()
        if self.energy_window:
            _min = self.get_min_energy_idx(energies)
            _max = self.get_max_energy_idx(energies)
            self.min_energy_idx = _min if _min else self.min_energy_idx
            self.max_energy_idx = _max if _max else self.max_energy_idx

        energies = energies[self.min_energy_idx: self.max_energy_idx + 1]

        return DosData(energies=energies,
                       total=np.array(self._total),
                       pdos=self._pdos,
                       vertical_lines=self.vertical_lines,
                       base_energy=self.base_energy)

    def get_max_energy_idx(self, energies):
        for i, e in enumerate(energies):
            if e > self.energy_window[1]:
                return i - 1

    def get_min_energy_idx(self, energies):
        for i, e in enumerate(energies):
            if e >= self.energy_window[0]:
                return i

    @property
    def _pdos(self):
        result = []
        for dos_by_site in self.complete_dos.pdos.values():
            pdos_kwargs = {}
            for orbital, dos_by_orbital in dos_by_site.items():
                pdos = np.array(
                    [dos_by_orbital[s] for s in [Spin.up, Spin.down]
                     if s in dos_by_orbital])

                pdos_kwargs[str(orbital)] = \
                    pdos[:, self.min_energy_idx:self.max_energy_idx + 1]
            try:
                result.append(PDos.from_dict(pdos_kwargs))
            except TypeError:
                logger.warning("Orbital DOSes are required. SET LORBIT = 11.")
                raise

        return result

    @property
    def _total(self):
        result = []
        for s in [Spin.up, Spin.down]:
            if s in self.complete_dos.densities:
                dos = self.complete_dos.densities[s]
                result.append(dos[self.min_energy_idx:self.max_energy_idx + 1])
        return result



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pathlib import Path

import pytest
from pymatgen.core import Composition
from pymatgen.io.vasp import Vasprun

from vise.analyzer.plot_band import BandEdgeForPlot, XTicks, BandMplSettings
from vise.analyzer.vasp.plot_band import greek_to_unicode, italic_to_roman, \
    BandPlotInfoFromVasp
from vise.analyzer.plot_band import BandMplPlotter
import numpy as np

try:
    import psutil
    PSUTIL_NOT_PRESENT = False
except ModuleNotFoundError:
    PSUTIL_NOT_PRESENT = True


def test_greek_to_unicode():
    assert greek_to_unicode("GAMMA") == "Γ"
    assert greek_to_unicode("SIGMA") == "Σ"
    assert greek_to_unicode("DELTA") == "Δ"
    assert greek_to_unicode("LAMBDA") == "Λ"


def test_italic_to_roman():
    assert italic_to_roman(r"S$\mid$$S_0$") == "S$\mid$${\\rm S}_0$"


test_data = [(True, None),
             (False, BandEdgeForPlot(vbm=-100, cbm=100,
                                     vbm_distances=[0], cbm_distances=[1, 2]))]


@pytest.mark.parametrize("is_metal,expected_band_edge", test_data)
def test_vasp_band_plotter(is_metal, expected_band_edge, mocker):
    mock_bs = mocker.MagicMock()
    mock_bs.efermi = 10
    mock_bs.is_metal.return_value = is_metal

    stub_vasprun = mocker.MagicMock()
    stub_vasprun.final_structure.composition = Composition("MgO2")
    stub_vasprun.get_band_structure.return_value = mock_bs

    energy = {"1": [np.array([[0.1], [0.2], [0.3]])]}
    distances = [np.array([0.0, 0.1, 0.2])]
    labels = ["A", "$A_0$", "GAMMA"]
    label_distances = [0.0, 0.1, 0.2]
    plot_data = {"energy": energy,
                 "distances": distances,
                 "vbm": [[0, -100]],
                 "cbm": [[1, 100], [2, 100]]}

    mock_bsp = mocker.patch("vise.analyzer.vasp.plot_band.BSPlotter",
                            unsafe=True)
    mock_bsp.return_value.bs_plot_data.return_value = plot_data
    mock_bsp.return_value.get_ticks_old.return_value = \
        {"label": labels, "distance": [0.0, 0.1, 0.2]}

    band_plot_info = BandPlotInfoFromVasp(stub_vasprun, "KPOINTS")
    plot_info = band_plot_info.make_band_plot_info()

    expected_x_ticks = XTicks(labels=["A", "${\\rm A}_0$", "Γ"],
                              distances=label_distances)

    assert plot_info.band_energy_infos["1"].band_energies == \
           [[[[0.1], [0.2], [0.3]]]]
    assert plot_info.band_energy_infos["1"].band_edge == expected_band_edge
    assert plot_info.distances_by_branch == [[0.0, 0.1, 0.2]]
    assert plot_info.x_ticks == expected_x_ticks
    assert plot_info.title == "MgO$_{2}$"


def test_draw_band_plotter_with_actual_vasp_files(test_data_files: Path):
    vasprun_file = str(test_data_files / "KO2_band_vasprun.xml")
    kpoints_file = str(test_data_files / "KO2_band_KPOINTS")
    vasprun = Vasprun(vasprun_file)
    plot_info = BandPlotInfoFromVasp(vasprun, kpoints_file).make_band_plot_info()
    plotter = BandMplPlotter(plot_info, [-10, 10])
    plotter.construct_plot()
    plotter.plt.show()


# def test_irrep_actual_vasp_files(test_data_files: Path):


def test_draw_two_bands(test_data_files: Path):
    mpl_settings = BandMplSettings(linewidth=[0.3, 1.0],
                                   circle_size=50,
                                   show_legend=True)

    vasprun_file = str(test_data_files / "CdAs2O6-vasprun1.xml")
    vasprun2_file = str(test_data_files / "CdAs2O6-vasprun2.xml")
    kpoints_file = str(test_data_files / "CdAs2O6-KPOINTS")
    vasprun = Vasprun(vasprun_file)
    vasprun2 = Vasprun(vasprun2_file)
    plot_info = BandPlotInfoFromVasp(
        vasprun, kpoints_file, first_band_plot_name="pbe",
        vasprun2=vasprun2, second_band_plot_name="dd hybrid",
        energy_window=[-20.0, 20.0]).make_band_plot_info()

    plotter = BandMplPlotter(plot_info, [-10, 10], mpl_defaults=mpl_settings)
    plotter.construct_plot()
    plotter.plt.show()


def test_energy_window(mocker):
    mock_bs = mocker.MagicMock()
    mock_bs.efermi = 0
    mock_bs.is_metal.return_value = True

    stub_vasprun = mocker.MagicMock()
    stub_vasprun.final_structure.composition = Composition("MgO2")
    stub_vasprun.get_band_structure.return_value = mock_bs

    energy = {"1": [np.array([[-0.2, -0.1, -0.3, -0.1],
                              [-0.2, -0.1, -0.3, 0.1],
                              [1.2, 1.3, 1.1, 1.2]])]}
    distances = [[0.0, 1.0]]
    labels = ["A", "GAMMA"]
    label_distances = [0.0, 1.0]
    plot_data = {"energy": energy,
                 "distances": distances,
                 "vbm": None,
                 "cbm": None}

    mock_bsp = mocker.patch("vise.analyzer.vasp.plot_band.BSPlotter",
                            unsafe=True)
    mock_bsp.return_value.bs_plot_data.return_value = plot_data
    mock_bsp.return_value.get_ticks_old.return_value = \
        {"label": labels, "distance": label_distances}

    plot_info = BandPlotInfoFromVasp(
        stub_vasprun, "KPOINTS", energy_window=[0.0, 1.0]).make_band_plot_info()

    assert (plot_info.band_energy_infos["1"].band_energies
            == [[[[-0.2, -0.1, -0.3, 0.1]]]])




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pathlib import Path

import numpy as np
import pytest
from pymatgen.electronic_structure.core import Spin

from vise.analyzer.band_edge_properties import (
    BandEdge, BandEdgeProperties, is_band_gap, merge_band_edge)
from vise.defaults import defaults
from vise.tests.helpers.assertion import assert_msonable, \
    assert_dataclass_almost_equal

parent_dir = Path(__file__).parent


actual_kpt = [[10.1, 10.2, 10.3], [10.4, 10.5, 10.6]]
expected_metal = \
    {'energies': 0.0, 'direct': None, 'transition': None}, None, None


def test_band_edge_msonable():
    assert_msonable(BandEdge(energy=0.0, spin=Spin.up, band_index=0,
                             kpoint_index=1, kpoint_coords=[0.0, 0.0, 0.0],
                             data_source="dos",
                             symbol="X1"))


def test_band_edge_equal():
    e1 = BandEdge(0.0, Spin.up,   band_index=0, kpoint_coords=[0.0, 0.0, 0.0])
    e2 = BandEdge(0.0, Spin.up,   band_index=0, kpoint_coords=[0.0, 0.0, 0.0])
    e3 = BandEdge(0.0, Spin.up,   band_index=0, kpoint_coords=[0.1, 0.0, 0.0])
    e4 = BandEdge(0.0, Spin.down, band_index=0, kpoint_coords=[0.0, 0.0, 0.0])
    assert e1.is_direct(e2) is True
    assert e1.is_direct(e3) is False
    assert e1.is_direct(e4) is False


def test_metal_judged_from_non_uniform_band_occupation():
    eigenvalues = {Spin.up: np.array([[0.0, 0.1, 0.2], [0.2, 0.3, 0.4]])}
    band_edge = BandEdgeProperties(eigenvalues=eigenvalues,
                                   nelect=4.0,
                                   magnetization=0.0,
                                   kpoint_coords=actual_kpt)

    assert band_edge.band_gap is None
    assert band_edge.is_direct is None
    assert band_edge.vbm_info is None
    assert band_edge.cbm_info is None


def test_metal_judged_from_fractional_nelect():
    eigenvalues = {Spin.up: np.array([[0.0, 1.0, 2.0], [0.1, 1.1, 2.1]])}
    integer_criterion = 0.1
    band_edge = BandEdgeProperties(eigenvalues=eigenvalues,
                                   nelect=4.0 + integer_criterion + 1e-5,
                                   magnetization=0.0,
                                   kpoint_coords=actual_kpt,
                                   integer_criterion=0.1)

    assert band_edge.band_gap is None
    assert band_edge.is_direct is None
    assert band_edge.vbm_info is None
    assert band_edge.cbm_info is None


def test_metal_judged_from_fractional_magnetization():
    eigenvalues = {Spin.up:   np.array([[0.0, 1.0, 10.0], [0.0, 1.1, 10.0]]),
                   Spin.down: np.array([[0.0, 1.4, 10.0], [0.0, 1.5, 10.0]])}
    integer_criterion = 0.1
    band_edge = BandEdgeProperties(eigenvalues=eigenvalues,
                                   nelect=3.0,
                                   magnetization=1.0 + integer_criterion + 1e-5,
                                   kpoint_coords=actual_kpt,
                                   integer_criterion=0.1)

    assert band_edge.band_gap is None
    assert band_edge.is_direct is None
    assert band_edge.vbm_info is None
    assert band_edge.cbm_info is None


@pytest.fixture
def nonmagnetic_band_edge():
    # k-point indices run fast.
    eigenvalues = {Spin.up: np.array([[0.0, 1.0, 2.0], [0.1, 1.1, 2.1]])}
    integer_criterion = 0.1
    return BandEdgeProperties(eigenvalues=eigenvalues,
                              nelect=4.0 + integer_criterion - 1e-5,
                              magnetization=0.0,
                              kpoint_coords=actual_kpt,
                              integer_criterion=0.1)


def test_nonmagnetic_insulator(nonmagnetic_band_edge):
    band_edge = nonmagnetic_band_edge
    assert pytest.approx(band_edge.band_gap) == 0.90

    assert band_edge.is_direct is False

    assert pytest.approx(band_edge.vbm_info.energy) == 1.1
    assert pytest.approx(band_edge.cbm_info.energy) == 2.0

    assert band_edge.vbm_info.spin == Spin.up
    assert band_edge.cbm_info.spin == Spin.up

    assert band_edge.vbm_info.band_index == 1
    assert band_edge.cbm_info.band_index == 2

    assert band_edge.vbm_info.kpoint_coords == [10.4, 10.5, 10.6]
    assert band_edge.cbm_info.kpoint_coords == [10.1, 10.2, 10.3]


@pytest.fixture
def magnetic_band_edge():
    eigenvalues = {Spin.up:   np.array([[0.0, 1.0, 10.0], [0.0, 1.1, 10.0]]),
                   Spin.down: np.array([[0.0, 1.4, 10.0], [0.0, 1.5, 10.0]])}
    return BandEdgeProperties(eigenvalues=eigenvalues,
                              nelect=3.0,
                              magnetization=1.0,
                              kpoint_coords=actual_kpt)


def test_magnetic_insulator(magnetic_band_edge):

    assert pytest.approx(magnetic_band_edge.band_gap) == 0.3

    assert magnetic_band_edge.is_direct is False

    assert pytest.approx(magnetic_band_edge.vbm_info.energy) == 1.1
    assert pytest.approx(magnetic_band_edge.cbm_info.energy) == 1.4

    assert magnetic_band_edge.vbm_info.spin == Spin.up
    assert magnetic_band_edge.cbm_info.spin == Spin.down

    assert magnetic_band_edge.vbm_info.band_index == 1
    assert magnetic_band_edge.cbm_info.band_index == 1

    assert magnetic_band_edge.vbm_info.kpoint_coords == [10.4, 10.5, 10.6]
    assert magnetic_band_edge.cbm_info.kpoint_coords == [10.1, 10.2, 10.3]


def test_min_gap(magnetic_band_edge):
    actual = magnetic_band_edge.min_gap_w_coords(same_energy_threshold=1e-5)
    expected = 1.4, [[10.4, 10.5, 10.6], [10.1, 10.2, 10.3]]
    assert actual == expected


def test_is_metal():
    band_edge_metal = BandEdgeProperties(
        eigenvalues={Spin.up: np.array([[0, 1, 2], [0, 3, 4]])},
        nelect=4.0,
        magnetization=0.0,
        kpoint_coords=actual_kpt)
    assert band_edge_metal.is_metal is True
    assert band_edge_metal.vbm_info is None
    assert band_edge_metal.cbm_info is None
    assert band_edge_metal.vbm_cbm is None
    assert band_edge_metal.band_gap is None
    assert repr(band_edge_metal) == "Metal"


def test_repr(magnetic_band_edge):
    expected = """Band gap 0.300 eV
VBM energy position: 1.1, spin:   up, band index 1, k-point index 1, k-point coords 10.400 10.500 10.600
CBM energy position: 1.4, spin: down, band index 1, k-point index 0, k-point coords 10.100 10.200 10.300"""
    assert repr(magnetic_band_edge) == expected


def test_is_band_gap():
    assert is_band_gap(None, [1.0, 1.0 + defaults.band_gap_criterion + 1e-5]) is True
    assert is_band_gap(None, [1.0, 1.0 + defaults.band_gap_criterion - 1e-5]) is False
    assert is_band_gap(None, None) is False


def test_merge_band_edge():
    edge1 = BandEdge(energy=0.0, spin=Spin.up, band_index=0,
                     kpoint_index=1, kpoint_coords=[0.0, 0.0, 0.0],
                     data_source="dos")
    edge2 = BandEdge(energy=-0.0011, spin=Spin.down, band_index=0,
                     kpoint_index=1, kpoint_coords=[0.1, 0.1, 0.1],
                     data_source="band")
    assert merge_band_edge(edge1, edge2, "vbm") == edge1
    assert merge_band_edge(edge1, edge2, "cbm") == edge2


def test_merge_band_edge_same():
    edge1 = BandEdge(energy=0.0, spin=Spin.up, band_index=0,
                     kpoint_index=1, kpoint_coords=[0.0, 0.0, 0.0],
                     data_source="dos")
    edge2 = BandEdge(energy=-0.09, spin=Spin.down, band_index=0,
                     kpoint_index=1, kpoint_coords=[0.1, 0.1, 0.1],
                     data_source="band")
    # Because the kpoint coords are different, edge2 is returned.
    actual = merge_band_edge(edge1, edge2, "vbm", threshold=0.1)
    assert_dataclass_almost_equal(actual, edge1)

    same_edge = BandEdge(energy=0.09, spin=Spin.up, band_index=0,
                         kpoint_index=1, kpoint_coords=[0.0, 0.0, 0.0],
                         data_source="band")
    # Higher energy is used for VBM.
    actual = merge_band_edge(edge1, same_edge, "vbm", threshold=0.1)
    expected = BandEdge(energy=0.09, spin=Spin.up, band_index=0,
                        kpoint_index=1, kpoint_coords=[0.0, 0.0, 0.0],
                        data_source="dos band")
    assert_dataclass_almost_equal(actual, expected)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import logging
import sys


def get_logger(name: str,
               level=logging.DEBUG,
               stream=sys.stdout,
               datetime: str = '%Y/%m/%d %H:%M:%S',
               logger_type: str = "simple",
               log_filename: str = None,
               file_handler_level: str = "WARNING"):

    if logger_type == "simple":
        log_format = "%(levelname)7s: %(message)s"
        formatter = logging.Formatter(log_format, datefmt=datetime)
    elif logger_type == "abundant":
        log_format = \
            "%(asctime)18s %(levelname)7s %(name)25s\n --> %(message)s"
        formatter = logging.Formatter(log_format, datefmt=datetime)
    else:
        raise ValueError

    logger = logging.getLogger(name)
    logger.setLevel(level)

    if log_filename:
        file_handler = logging.FileHandler(log_filename)
        file_handler.setFormatter(formatter)
        file_handler.setLevel(file_handler_level)
        logger.addHandler(file_handler)

    stream_handler = logging.StreamHandler(stream=stream)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    return logger



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from vise.input_set.fft_grids import vasp_grid


def test_vasp_grids():
    actual = vasp_grid(encut=250, lattice_length=10.0, symprec="Normal")
    assert actual == 39


